     module matrix_manipulations
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!this module contains routines that facilitate manipulating CSR matrices. 
! calls the csr libraries.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! VERSION: 1.0
! WRITTEN BY: NOVELTIS (A.KLONECKI)
! CREATION DATE: 26/09/2005
! REVISION DATE:

contains

     subroutine multiply_matrices_csr( iw             & !work array
                             ,Ha_csr                  & !matrix A 
                             ,Hb_csr                  & !matrix B 
                             ,Hc_csr                  & !matrix C=A*B
                             ,nmaxc                   &
                             )
                           
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! routine that multiply two matrices in CSR format, it uses the routine amub
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      use floats
      use MATRIX_CSR_STRUCTURE


      implicit none 
!
!INPUT:
      integer                                :: dim1
      integer, dimension(:)                  :: iw
      type(CSR_FORMAT)                       :: Ha_csr
      type(CSR_FORMAT)                       :: Hb_csr
      integer                                :: nmaxc
!OUTPUT:
      type(CSR_FORMAT)                       :: Hc_csr
!LOCAL:
      integer*4        :: job,ierr,np,NPRF,ALLOC_ERR


      job = 1 
! the problem with the iw variable is that it is dimensioned by the number of 
! columns of the full matrix which is large: 1522800 for 90*180*47*2
! To avoid intializing iw to zero in amub (as coded in SPARSKIT, and which takes time),
! the initialization is done only once in this routine. The variable iw is reinitialized to 
! zero in amub2 uniquely for the indicies which were modified.  

      if (associated(Hc_csr%H)) then
        call deallocate_operator(Hc_csr)
      end if

      NPRF=Ha_csr%NPRF

      Hc_csr%NPRF  = NPRF
      Hc_csr%nrow  = Ha_csr%nrow
      Hc_csr%ncol  = Hb_csr%ncol
      Hc_csr%nmax  = nmaxc
!
      allocate(Hc_csr%H(nmaxc,NPRF),STAT = ALLOC_ERR)
      call check_allocate(ALLOC_ERR,"multiply_matrices_csr (")
      allocate(Hc_csr%j(nmaxc,NPRF),STAT = ALLOC_ERR)
      call check_allocate(ALLOC_ERR,"multiply_matrices_csr (")
      allocate(Hc_csr%i(Hc_csr%nrow+1,NPRF),STAT = ALLOC_ERR)
      call check_allocate(ALLOC_ERR,"multiply_matrices_csr (")
!
!      Hc_csr%H=0.
!      Hc_csr%j=0
!      Hc_csr%i=0
      do np=1,NPRF

!amub2 is a SPARSKIT routine modified by NOVELTIS to run faster. For details
!see the routine amub2 in this file. 
          call amub2(Ha_csr%nrow,Hb_csr%ncol,job,              &
                 Ha_csr%H(:,np),Ha_csr%j(:,np),Ha_csr%i(:,np), &    !matrix A
                 Hb_csr%H(:,np),Hb_csr%j(:,np),Hb_csr%i(:,np), &    !matrix B
                 Hc_csr%H(:,np),Hc_csr%j(:,np),Hc_csr%i(:,np), &    !out:A*B
                 nmaxc,iw,ierr) 
          call check_error_sparse(ierr,"amub")
      end do

      return
      end subroutine multiply_matrices_csr
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     subroutine multiply_matrix_csr_vector(           &
!  input
                              H_csr                   & !matrix in format csr
                             ,vector                  & !vector
                             ,nrow                    & !dim 1
                             ,NPRF                    & !dim 2
!  output
                             ,vector_out              & !the resulting vector
                             )

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! routine that multiplies a matrix in CSR format by a vector, uses the routine amux
! from SPARSKIT
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      use floats
      use MATRIX_CSR_STRUCTURE


      implicit none
!
!INPUT:
      type(CSR_FORMAT)                       :: H_csr
      real(std),dimension(:)                 :: vector
      integer                                :: nrow
      integer                                :: NPRF
!OUTPUT:
      real(std),dimension(nrow,NPRF)         :: vector_out
!LOCAL:
      integer*4        :: np,ALLOC_ERR,ib_i,ib_k

      vector_out=0.
      DO np=1,NPRF
	DO ib_i = 1,H_csr%nrow
	  DO ib_k = H_csr%i(ib_i,np), H_csr%i(ib_i+1,np)-1
	    vector_out(ib_i,np) = vector_out(ib_i,np) &
	       + H_csr%H(ib_k,np)*vector(H_csr%j(ib_k,np))
	  END DO
	END DO
      END DO

      return
      end subroutine multiply_matrix_csr_vector

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     subroutine multiply_matrixT_csr_vector(           & 
!  input
                              H_csr                   & !matrix in format csr
                             ,vector                  & !vector
                             ,nrow                    & !dim 1
                             ,NPRF                    & !dim 2
!  output
                             ,vector_out              & !the resulting vector
                             )

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! routine that multiplies the transpose of a matrix in CSR format by a vector, 
! uses the routine atmux
! from SPARSKIT
! written by Andrea Piacentini
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      use floats
      use MATRIX_CSR_STRUCTURE


      implicit none
!
!INPUT:
      type(CSR_FORMAT)                       :: H_csr
      real(std),dimension(:)                 :: vector
      integer                                :: nrow
      integer                                :: NPRF
!OUTPUT:
      real(std),dimension(nrow,NPRF)         :: vector_out
!LOCAL:
      integer*4        :: np,ALLOC_ERR

      do np=1,NPRF
         call atmux(H_csr%nrow,vector,vector_out(:,np),           &
                    H_csr%H(:,np),H_csr%j(:,np),H_csr%i(:,np))
      end do
            
      return
      end subroutine multiply_matrixT_csr_vector

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      SUBROUTINE addmult_matrixT_csr_vector(           & 
!  input
                              H_csr                   & !matrix in format csr
                             ,vector                  & !vector
                             ,ncol                    & !dim 1
                             ,NPRF                    & !dim 2
!  output
                             ,vector_out              & !the resulting vector
                             )

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! routine that multiplies the transpose of a matrix in CSR format by a vector, 
! uses the routine atmux
! from SPARSKIT
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      use floats
      use MATRIX_CSR_STRUCTURE


      implicit none
!
!INPUT:
      type(CSR_FORMAT)                       :: H_csr
      real(std),dimension(ncol,NPRF)         :: vector
      integer                                :: ncol
      integer                                :: NPRF
!OUTPUT:
      real(std),dimension(:)         :: vector_out
!LOCAL:
      integer*4        :: np,ALLOC_ERR,ib_i,ib_k

      vector_out=0.
      DO np=1,NPRF
	DO ib_i = 1,H_csr%nrow
	  DO ib_k = H_csr%i(ib_i,np), H_csr%i(ib_i+1,np)-1
	    vector_out(H_csr%j(ib_k,np)) = vector_out(H_csr%j(ib_k,np)) &
	       + vector(ib_i,np)*H_csr%H(ib_k,np)
	  END DO
	END DO
      END DO

      return
      end subroutine addmult_matrixT_csr_vector

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


      subroutine copy_matrix_csr(Ha_csr    &
                                ,Hb_csr) 
     
      use floats
      use MATRIX_CSR_STRUCTURE

      implicit none
!
!INPUT:
      type(CSR_FORMAT)                       :: Ha_csr
! OUTPUT
      type(CSR_FORMAT)                       :: Hb_csr
! LOCAL
      integer*4  :: i,np

      if (associated(Hb_csr%H)) then
        call deallocate_operator(Hb_csr)
      end if
!
      Hb_csr%nrow=Ha_csr%nrow
      Hb_csr%ncol=Ha_csr%ncol
      Hb_csr%nmax=Ha_csr%nmax
      Hb_csr%NPRF=Ha_csr%NPRF
      allocate(Hb_csr%H(Hb_csr%nmax,Hb_csr%NPRF))
      allocate(Hb_csr%j(Hb_csr%nmax,Hb_csr%NPRF))
      allocate(Hb_csr%i(Hb_csr%nrow+1,Hb_csr%NPRF))
!
      do np=1,Ha_csr%NPRF
        do i=1,Hb_csr%nmax
          Hb_csr%H(i,np)=Ha_csr%H(i,np)
          Hb_csr%j(i,np)=Ha_csr%j(i,np)
        end do
        do i=1,Hb_csr%nrow+1
          Hb_csr%i(i,np)=Ha_csr%i(i,np)
        end do
      end do
      return
      end subroutine copy_matrix_csr
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! SPARSKIT routine modified slightly to run faster. The intialization 
! of the array iw is very time consuming for arrays with large number 
! of columns. Instead of intializing the array iw each time the routine
! is called, the inialization is done once in the calling routine. The 
! indicies used by amub2 are reset to zero at the end of the routine. 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
       subroutine amub2 (nrow,ncol,job,a,ja,ia,b,jb,ib,  &
                        c,jc,ic,nzmax,iw,ierr) 
      real*8 a(*), b(*), c(*) 
      integer ja(*),jb(*),jc(*),ia(nrow+1),ib(*),ic(*),iw(ncol)
      real*8 scal 
      logical values
      values = (job .ne. 0) 
      len = 0
      ic(1) = 1 
      ierr = 0

!      do 1 j=1, ncol
!         iw(j) = 0
! 1    continue
!c
      do 500 ii=1, nrow 
!c     row i 
         do 200 ka=ia(ii), ia(ii+1)-1 
	    if (values) scal = a(ka)
	    jj   = ja(ka)
	    do 100 kb=ib(jj),ib(jj+1)-1
               jcol = jb(kb)
               jpos = iw(jcol)
               if (jpos .eq. 0) then
                  len = len+1
                  if (len .gt. nzmax) then
                     ierr = ii
                     return
                  endif
                  jc(len) = jcol
                  iw(jcol)= len
                  if (values) c(len)  = scal*b(kb)
               else
                  if (values) c(jpos) = c(jpos) + scal*b(kb)
               endif
 100	    continue
 200     continue
         do 201 k=ic(ii), len
	    iw(jc(k)) = 0
 201     continue
         ic(ii+1) = len+1
 500  continue
      return
      end subroutine amub2
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
       subroutine check_error_sparse(ierr,str_routine)
       implicit none

       integer,intent(in)        ::ierr
       character(*),intent(in)   ::str_routine

       integer                   :: ltrim

       if (ierr /= 0) then
         write(*,*) 'Pb in the sparse matrix routine ',str_routine(1:ltrim(str_routine))
         write(*,*) 'error number ierr=',ierr,' exiting...'
         write(*,*) 'check if the number of non-zero elements in the result matrix not too small'
         call exit(-1)
       end if
       return
       end subroutine check_error_sparse
!!!!!!!!!!!!!!!!

       subroutine check_allocate(ierr,str_routine)
       implicit none

       integer,intent(in)        ::ierr
       character(*),intent(in)   ::str_routine

       integer                   :: ltrim

       if (ierr /= 0) then
         write(*,*) 'Pb while allocating memory in routine ',str_routine(1:ltrim(str_routine))
         write(*,*) 'error ierr= ',ierr
         write(*,*) 'Check if not out of memory, exiting...'
         call exit(-1)
       end if
       return
       end subroutine check_allocate



      subroutine deallocate_operator(H_csr)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! THIS ROUTINE DEALLOCATES THE ARRAYS IN THE CSR_FORMAT structure
      use MATRIX_CSR_STRUCTURE

      implicit none 
!
!INPUT:
      type(CSR_FORMAT)                       :: H_csr
      integer                                :: il_err

      deallocate(H_csr%H,STAT=il_err)
      deallocate(H_csr%j,STAT=il_err)
      deallocate(H_csr%i,STAT=il_err)

      end subroutine deallocate_operator

      end module matrix_manipulations


