!
!******************************************************************
!******************************************************************
! VERSION: 1.1
! WRITTEN BY: NOVELTIS (A.KLONECKI)
! THE PROGRAM IS BASED ON THE ALGORITHMS PRESENT IN THE ASSIMILATION CODE
! WRITTEN BY JF Lamarque 
! 
! CREATION DATE: 10/08/2005
! REVISION DATE: 12/01/2006
!    -added flags : flag_averaging_kernel,flag_temp_interp
!    -handles non-constant observed pressure levels
!    -instead of estimating the number of nonzero elements for the products 
!     of multiplication of two matrices (nmax), these values are computed 
!     based on analytic expressions.

!******************************************************************

  subroutine observ_operator  (                 &
! INPUT MODEL DATA:
                            nlat            & !dimensions:
                           ,nlon            &
                           ,nlev_m          &
                           ,ntime           &
                           ,latgrid         & !data:
                           ,longrid         & 
                           ,pres_m          &
                           ,date_MOCAGE     &
! INPUT OBS DATA:
                           ,NPRF            & !dimensions:
                           ,n_lvls          &
                           ,n_lvls_out   &
                           ,lat             & !data:
                           ,lon             &
                           ,date            &
                           ,pres_o          &
! INPUT COMPLEMENTARY DATA:
                           ,avg_kernel      &
! EXECUTION OPTIONS:
                           ,flag_averaging_kernel &
                           ,flag_temp_interp      &
! OUTPUT DATA:
                           ,Hres_csr        & !Observational Op. in CSR format
                           ,discard_obs     &
                                            ) !
!
  use floats
  use MATRIX_CSR_STRUCTURE
  use transformations_matrix
  use matrix_manipulations
  implicit none
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! DECLARATIONS:
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! INPUT DATA:
 !numb. of model longitudes and latitudes
  integer*4,intent(in)                      :: nlon,nlat  
!numb. of model veritcal levels
  integer*4,intent(in)                      :: nlev_m   
!numb. of temporary MOCAGE files: if 1 no temporal interpolation  
!                                 if 2 - temporal interpolation :
  integer*4,intent(in)                      :: ntime     
                           
! model lats and lons:
  real(std), dimension(nlat,nlon),intent(in):: latgrid
  real(std), dimension(nlat,nlon),intent(in):: longrid
! model fields:
  real(std), dimension(nlon,nlat,nlev_m,ntime),intent(in) :: pres_m
  real(std), dimension(ntime),intent(in)    :: date_MOCAGE

!numb. of obser. profiles in MOCAGE interval being treated:
  integer,intent(in)                        :: NPRF       
!maximum numb. of obs vertical levels
  integer*4,intent(in)                      :: n_lvls  
!this variable dimensions the AvKernel(n_lvls_out,n_lvls),=1 in case
! of integrated column data   
  integer*4,intent(in)                      :: n_lvls_out 
                           
  real(std),dimension(NPRF),intent(in)      :: lat      !IASI lat
  real(std),dimension(NPRF),intent(in)      :: lon      !IASI lon
  real(std),dimension(NPRF),intent(in)      :: date     !IASI dates

  real(std),dimension(n_lvls,NPRF),intent(in):: pres_o

  real(std), dimension(n_lvls_out,n_lvls,NPRF),intent(in):: avg_kernel

  logical,intent(in)                      :: flag_temp_interp
  logical,intent(in)                      :: flag_averaging_kernel

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! OUTPUT DATA:
  TYPE(CSR_FORMAT),intent(out)              :: Hres_csr
  logical, dimension(NPRF),intent(out)      :: discard_obs

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! LOCAL VARIABLES:
  integer, pointer, dimension(:)        :: iw  ! work array for the routine amub
! THE FOUR MATRICES CORRESPONDING TO THE 4 OPERATORS, MATRICES IN CSR FORMAT:
  type(CSR_FORMAT)                          :: Hh_csr  !horizontal operator
  type(CSR_FORMAT)                          :: Hv_csr  !vertical operator
  type(CSR_FORMAT)                          :: Ht_csr  !temporal operator
  type(CSR_FORMAT)                          :: Ha_csr  !smoothing operator
! TEMPORARY MATRICES FOR STORING THE RESULTS OF MULTIPLICATIONS OF MATRICES:
  type(CSR_FORMAT)                          :: Hres1_csr  !result of Hv*Hh
  type(CSR_FORMAT)                          :: Hres2_csr  !result of Ht*(Hv*Hh)


! CONTAINS THE STATE VECTOR with PRESSURE data:
  real(std),pointer,dimension(:)        :: vector
! MODEL PRESSURE INTEGRATED TO THE LOCATION OF OBSERVATION 
  real(std),pointer,dimension(:,:)      :: pres_inth
! FUNCTION:
  integer*4                                 :: get_number_levels

!# of model levels used in vertical interpolation, if the averaging kernel is used, 
! this number is used to store the  number of non-zero elements in the final
!  matrix  
  integer*4                                 :: lvls_out 
  integer*4                                 :: np   !indexing variable
! number of non-zero elements in the CSR format:
  integer*4                                 :: nmax 
  integer*4                                 :: ierr !error index

  integer*4                                 :: ALLOC_ERR


!******************************************************************
! END OF DECLARATIONS
 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! 1) PREPARE THE MATRIX TO INTERPOLATE MODEL CO AND PRES FROM MODEL GRID TO 
!    OBS LOCATIONS
          call transf_horiz_interp_matrix(NPRF,nlon,nlat,nlev_m,ntime,         &
                                       latgrid,longrid,                        &
                                       lat,lon,  &
                                       discard_obs,             &
                                       Hh_csr)
!!$            WRITE(14871,*) 'Hh_csr stats min max'
!!$            WRITE(14871,*) MINVAL(Hh_csr%H)
!!$            WRITE(14871,*) MAXVAL(Hh_csr%H)
!!$
!!$            WRITE(*,*) 'Hh_csr stats min max'
!!$            WRITE(*,*) MINVAL(Hh_csr%H)
!!$            WRITE(*,*) MAXVAL(Hh_csr%H)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! 2) PREPARE THE MATRIX TO INTERPOLATE TO OBS VERTICAL LEVELS (coarse): 
!
! need the interpolated pressure levels for the vertical interpolation:
          allocate(vector(nlat*nlon*nlev_m*ntime))
          call create_state_vector(nlat,nlon,nlev_m,ntime,pres_m,vector)

          allocate(pres_inth(Hh_csr%nrow,NPRF),STAT = ALLOC_ERR)
          call check_allocate(ALLOC_ERR,"multiply_matrix_csr_vector")
          call multiply_matrix_csr_vector(Hh_csr,vector,          &
                                          Hh_csr%nrow,NPRF,pres_inth)
          deallocate(vector)                                     

! pres_inth now contains the model pressures on the obs horizontal grid
          call transf_vert_interp_matrix(pres_inth,pres_o,    &
                                discard_obs,     &
                                NPRF,nlev_m,n_lvls,ntime,       &
                                Hv_csr)
                                
!!$            WRITE(14871,*) 'Hv_csr stats min max'
!!$            WRITE(14871,*) MINVAL(Hv_csr%H)
!!$            WRITE(14871,*) MAXVAL(Hv_csr%H)
!!$            WRITE(*,*) 'Hv_csr stats min max'
!!$            WRITE(*,*) MINVAL(Hv_csr%H)
!!$            WRITE(*,*) MAXVAL(Hv_csr%H)

          deallocate(pres_inth)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! 3) PREPARE THE MATRIX TO INTERPOLATE WITH RESPECT TO TIME 

          if (flag_temp_interp) then   !if the temporal interpolation is 
                                            ! necessary:

            call transf_temp_interp_matrix(n_lvls,                 &
                          NPRF,ntime,                              & 
                          date_MOCAGE(1:2),                        &
                          date,                     &
                          Ht_csr)
!!$            WRITE(14871,*) 'Ht_csr stats min max'
!!$            WRITE(14871,*) MINVAL(Ht_csr%H)
!!$            WRITE(14871,*) MAXVAL(Ht_csr%H)
!!$
!!$            WRITE(*,*) 'Ht_csr stats min max'
!!$            WRITE(*,*) MINVAL(Ht_csr%H)
!!$            WRITE(*,*) MAXVAL(Ht_csr%H)
          end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! 4) the Averaging matrix is already calculated, convert to the CSR format
          if (flag_averaging_kernel) then 
            Ha_csr%NPRF = NPRF
            Ha_csr%ncol = n_lvls
            Ha_csr%nrow = n_lvls_out
            nmax=Ha_csr%ncol*Ha_csr%nrow
            Ha_csr%nmax = nmax
            allocate(Ha_csr%H(nmax,NPRF))    !for now the same for all profiles
            allocate(Ha_csr%j(nmax,NPRF))    
            allocate(Ha_csr%i(Ha_csr%nrow+1,NPRF)) 
	    do np=1,NPRF
	      CALL dnscsr(Ha_csr%nrow,Ha_csr%ncol,nmax,  &
		 avg_kernel(:,:,np),Ha_csr%nrow,Ha_csr%H(:,np),  &
		 Ha_csr%j(:,np),Ha_csr%i(:,np),ierr)
            end do
          end if

!---
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! FINISHED CONSTRUCTING THE MATRICES, NOW MULTIPLY THEM TO OBTAIN THE FINAL 
! OBSERVATIONAL OPERATOR H=Ha*Ht*Hv*Hh
! FOR THE INDIVIDUAL MATRICES IT WAS STRAIGHTFORWARD TO FIND THE NUMBER OF
! NON ZERO ELEMENTS. FOR THE PRODUCTS OF TWO MATRICES, THE NUMBER OF NON ZERO
! ELEMENTS CAN BE ALSO ESTIMATED ANALYTICALLY. THE FORMULAS ARE GIVEN 
! BELOW.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! the multplication of Hv*Hh gives 8 non zero elements in each row.
          nmax = 8*Hv_csr%nrow

          allocate(iw(Hh_csr%ncol))           !work array needed for routine amub
          iw=0

          call multiply_matrices_csr(iw &
             ,Hv_csr    & !matrix A
             ,Hh_csr    & !matrix B
             ,Hres1_csr & !out:A*B
             ,nmax)   
!!$            WRITE(14871,*) 'Hres1_csr stats min max'
!!$            WRITE(14871,*) MINVAL(Hres1_csr%H)
!!$            WRITE(14871,*) MAXVAL(Hres1_csr%H)
!!$
!!$            WRITE(*,*) 'Hres1_csr stats min max'
!!$            WRITE(*,*) MINVAL(Hres1_csr%H)
!!$            WRITE(*,*) MAXVAL(Hres1_csr%H)

          call deallocate_operator(Hh_csr)
          deallocate(iw)

! before dealocating Hv_csr determine how many model levels are used in 
! the vertical interpolation. This number is needed later to dimension the
! final csr array with the operator H (if the averaging kernel is used). 
          if ( flag_averaging_kernel ) then 
            lvls_out=get_number_levels(Hv_csr,discard_obs,NPRF,n_lvls)
          end if
!
          call deallocate_operator(Hv_csr)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!                
!next calculate: H_t*H_res where H_res is H_v*H_h:
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!                
          if (flag_temp_interp) then      
!the multiplication of Ht*Hres gives 16 non zero values in each row
            nmax = 16 * Ht_csr%nrow
                               
            allocate(iw(Hres1_csr%ncol))    !work array needed for routine amub
            iw=0
            call multiply_matrices_csr(iw     &
              ,Ht_csr     &    !matrix A
              ,Hres1_csr  &    !matrix B
              ,Hres2_csr  &    !out:A*B
              ,nmax)
            call deallocate_operator(Ht_csr)
            deallocate(iw)
          else
            call copy_matrix_csr(Hres1_csr,Hres2_csr)
          end if  !if flag_temp_interp
!!$            WRITE(14871,*) 'Hres2_csr stats min max'
!!$            WRITE(14871,*) MINVAL(Hres2_csr%H)
!!$            WRITE(14871,*) MAXVAL(Hres2_csr%H)
!!$
!!$            WRITE(*,*) 'Hres2_csr stats min max'
!!$            WRITE(*,*) MINVAL(Hres2_csr%H)
!!$            WRITE(*,*) MAXVAL(Hres2_csr%H)

          call deallocate_operator(Hres1_csr)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!              
          if ( flag_averaging_kernel ) then
! and finaly calculate H_a*H_res where H_res is : H_t*H_v*H_h
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!                
            nmax = lvls_out * 16 * Ha_csr%nrow
                               
            allocate(iw(Hres2_csr%ncol))    !work array needed for routine amub
            iw=0
            call multiply_matrices_csr(iw      &
              ,Ha_csr        &    !matrix A
              ,Hres2_csr     &    !matrix B
              ,Hres_csr      &    !out:A*B
              ,nmax)
            call deallocate_operator(Ha_csr)
            deallocate(iw)
          else
            call copy_matrix_csr(Hres2_csr,Hres_csr)
          end if
!!$            WRITE(14871,*) 'Inner Hres_csr stats min max'
!!$            WRITE(14871,*) MINVAL(Hres_csr%H)
!!$            WRITE(14871,*) MAXVAL(Hres_csr%H)
!!$
!!$            WRITE(*,*) 'Inner Hres_csr stats min max'
!!$            WRITE(*,*) MINVAL(Hres_csr%H)
!!$            WRITE(*,*) MAXVAL(Hres_csr%H)
          call deallocate_operator(Hres2_csr)
!                                    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!                
! finished calculating the H matrix,
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

      return
      end
