!
!******************************************************************
!******************************************************************
! VERSION: 1.1
! WRITTEN BY: NOVELTIS (A.KLONECKI)
! THE PROGRAM IS BASED ON THE ALGORITHMS PRESENT IN THE ASSIMILATION CODE
! WRITTEN BY JF Lamarque 
! 
! CREATION DATE: 10/08/2005
! REVISION DATE: 12/01/2006
!    -added flags : flag_averaging_kernel,flag_temp_interp
!    -handles non-constant observed pressure levels
!    -instead of estimating the number of nonzero elements for the products 
!     of multiplication of two matrices (nmax), these values are computed 
!     based on analytic expressions.

!******************************************************************

SUBROUTINE observ_operator_tc  (                 &
! INPUT MODEL DATA:
                            nlat            & !dimensions:
                           ,nlon            &
                           ,nlev_m          &
                           ,ntime           &
                           ,latgrid         & !data:
                           ,longrid         & 
                           ,pres_m          &
                           ,pres_m_int      &
                           ,date_MOCAGE     &
! INPUT OBS DATA:
                           ,NPRF            & !dimensions:
                           ,n_lvls_out      &
                           ,lat             & !data:
                           ,lon             &
                           ,date            &
                           ,pres_bounds     &
! EXECUTION OPTIONS:
                           ,flag_temp_interp      &
! OUTPUT DATA:
                           ,Hres_csr        & !Observational Op. in CSR format
                           ,discard_obs     &
                                            ) !
!
  USE floats
  USE MATRIX_CSR_STRUCTURE
  USE transformations_matrix
  USE matrix_manipulations
  IMPLICIT NONE
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! DECLARATIONS:
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! INPUT DATA:
!numb. of model longitudes and latitudes
  INTEGER*4,INTENT(in)                      :: nlon,nlat  
!numb. of model veritcal levels
  INTEGER*4,INTENT(in)                      :: nlev_m   
!numb. of temporary MOCAGE files: if 1 no temporal interpolation  
!                                 if 2 - temporal interpolation :
  INTEGER*4,INTENT(in)                      :: ntime     
                           
! model lats and lons:
  REAL(std), DIMENSION(nlat,nlon),INTENT(in):: latgrid
  REAL(std), DIMENSION(nlat,nlon),INTENT(in):: longrid
! model fields:
  REAL(std), DIMENSION(nlon,nlat,nlev_m,ntime),INTENT(in) :: pres_m
  REAL(std), DIMENSION(nlon,nlat,nlev_m,ntime),INTENT(in) :: pres_m_int
  REAL(std), DIMENSION(ntime),INTENT(in)    :: date_MOCAGE

!numb. of obser. profiles in MOCAGE interval being treated:
  INTEGER,INTENT(in)                        :: NPRF       
!this variable dimensions the AvKernel(n_lvls_out,n_lvls),=1 in case
! of integrated column data   
  INTEGER*4,INTENT(in)                      :: n_lvls_out 
                           
  REAL(std),DIMENSION(NPRF),INTENT(in)      :: lat      !IASI lat
  REAL(std),DIMENSION(NPRF),INTENT(in)      :: lon      !IASI lon
  REAL(std),DIMENSION(NPRF),INTENT(in)      :: date     !IASI dates
  REAL(std),DIMENSION(2,n_lvls_out,NPRF),INTENT(in) :: pres_bounds !col. bounds

  logical,intent(in)                      :: flag_temp_interp

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! OUTPUT DATA:
  TYPE(CSR_FORMAT),intent(out)              :: Hres_csr
  LOGICAL, DIMENSION(NPRF),INTENT(out)      :: discard_obs

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! LOCAL VARIABLES:
  INTEGER, POINTER, DIMENSION(:)        :: iw  ! work array for the routine amub
! THE FOUR MATRICES CORRESPONDING TO THE 4 OPERATORS, MATRICES IN CSR FORMAT:
  TYPE(CSR_FORMAT)                          :: Hh_csr  !horizontal operator
  TYPE(CSR_FORMAT)                          :: Ht_csr  !temporal operator
  TYPE(CSR_FORMAT)                          :: Ha_csr  !smoothing operator
! TEMPORARY MATRICES FOR STORING THE RESULTS OF MULTIPLICATIONS OF MATRICES:
  TYPE(CSR_FORMAT)                          :: Hres2_csr  !result of Ht*(Hv*Hh)


! CONTAINS THE STATE VECTOR with PRESSURE data:
  REAL(std),POINTER,DIMENSION(:)        :: vector
! MODEL PRESSURE INTEGRATED TO THE LOCATION OF OBSERVATION 
  REAL(std),POINTER,DIMENSION(:,:)      :: pres_inth
  REAL(std),POINTER,DIMENSION(:,:)      :: pres_inth_int

  REAL(std), DIMENSION(n_lvls_out,nlev_m,NPRF)  :: avg_kernel
!maximum numb. of obs vertical levels
  INTEGER*4                                 :: n_lvls  

!# of model levels used in vertical interpolation, 
! if the averaging kernel is used, 
! this number is used to store the  number of non-zero elements in the final
!  matrix  
  INTEGER*4                                 :: lvls_out 
  INTEGER*4                                 :: np   !indexing variable
! number of non-zero elements in the CSR format:
  INTEGER*4                                 :: nmax 
  INTEGER*4                                 :: ierr !error index

  INTEGER*4                                 :: ALLOC_ERR


!******************************************************************
! END OF DECLARATIONS
 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! 1) PREPARE THE MATRIX TO INTERPOLATE MODEL CO AND PRES FROM MODEL GRID TO 
!    OBS LOCATIONS
  CALL transf_horiz_interp_matrix(NPRF,nlon,nlat,nlev_m,ntime,        &
                                  latgrid,longrid,                    &
                                  lat,lon,                            &
                                  discard_obs,                        &
                                  Hh_csr)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! 2) NO NEED OF VERTICAL INTERPOLATION: 
!
  n_lvls = nlev_m
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! 3) PREPARE THE MATRIX TO INTERPOLATE WITH RESPECT TO TIME 

  IF (flag_temp_interp) THEN   !if the temporal interpolation is 
                               ! necessary:

      CALL transf_temp_interp_matrix(n_lvls,                &
                                     NPRF,ntime,            & 
                                     date_MOCAGE(1:2),      &
				     date,                  &
				     Ht_csr)
  END IF
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! FINISHED CONSTRUCTING THE MATRICES, NOW MULTIPLY THEM TO OBTAIN THE FINAL 
! OBSERVATIONAL OPERATOR H=Ha*Ht*Hh
! FOR THE INDIVIDUAL MATRICES IT WAS STRAIGHTFORWARD TO FIND THE NUMBER OF
! NON ZERO ELEMENTS. FOR THE PRODUCTS OF TWO MATRICES, THE NUMBER OF NON ZERO
! ELEMENTS CAN BE ALSO ESTIMATED ANALYTICALLY. THE FORMULAS ARE GIVEN 
! BELOW.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!                
! calculate: H_t*H_h:
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!                
  IF (flag_temp_interp) THEN      
!the multiplication of Ht*Hres gives 16 non zero values in each row
      nmax = 16 * Ht_csr%nrow
                               
      ALLOCATE(iw(Hh_csr%ncol))    !work array needed for routine amub
      iw=0
      CALL multiply_matrices_csr(iw,         &
	                         Ht_csr,     &    !matrix A
				 Hh_csr,     &    !matrix B
				 Hres2_csr,  &    !out:A*B
				 nmax)
      CALL deallocate_operator(Ht_csr)
      DEALLOCATE(iw)
  ELSE
      CALL copy_matrix_csr(Hh_csr,Hres2_csr)
  END IF!if flag_temp_interp
!
  CALL deallocate_operator(Hh_csr)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! need the interpolated pressure levels for the total column:
  ALLOCATE(vector(nlat*nlon*nlev_m*ntime))
  CALL create_state_vector(nlat,nlon,nlev_m,ntime,pres_m,vector) 
  ALLOCATE(pres_inth(Hres2_csr%nrow,NPRF),STAT = ALLOC_ERR)
  CALL check_allocate(ALLOC_ERR,"multiply_matrix_csr_vector")
  CALL multiply_matrix_csr_vector(Hres2_csr,vector,          &
                                  Hres2_csr%nrow,NPRF,pres_inth)
  DEALLOCATE(vector)
  NULLIFY(vector)
!
  ALLOCATE(vector(nlat*nlon*nlev_m*ntime))
  CALL create_state_vector(nlat,nlon,nlev_m,ntime,pres_m_int,vector)
  ALLOCATE(pres_inth_int(Hres2_csr%nrow,NPRF),STAT = ALLOC_ERR)
  CALL check_allocate(ALLOC_ERR,"multiply_matrix_csr_vector")
  CALL multiply_matrix_csr_vector(Hres2_csr,vector,          &
                                  Hres2_csr%nrow,NPRF,pres_inth_int)
  DEALLOCATE(vector)
  NULLIFY(vector)
!
! pres_inth now contains the model pressures on the obs horizontal grid
! pres_inth_int now contains the model interface press. on the obs hor. grid
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! 4) Compute the avg kernel matrix for total columns (cubic), 
!    then convert it to the CSR format
!
  CALL transf_totcol_avg_kernel(nlev_m,n_lvls_out,NPRF,pres_inth, &
                                pres_inth_int,pres_bounds,avg_kernel)
!
  DEALLOCATE(pres_inth)
  DEALLOCATE(pres_inth_int)
!
  Ha_csr%NPRF = NPRF
  Ha_csr%ncol = n_lvls
  Ha_csr%nrow = n_lvls_out
  nmax=Ha_csr%ncol*Ha_csr%nrow
  Ha_csr%nmax = nmax
  ALLOCATE(Ha_csr%H(nmax,NPRF))    !for now the same for all profiles
  ALLOCATE(Ha_csr%j(nmax,NPRF))    
  ALLOCATE(Ha_csr%i(Ha_csr%nrow+1,NPRF))   
  DO np=1,NPRF
    CALL dnscsr(Ha_csr%nrow,Ha_csr%ncol,nmax,  &
       avg_kernel(:,:,np),Ha_csr%nrow,Ha_csr%H(:,np),  &
       Ha_csr%j(:,np),Ha_csr%i(:,np),ierr)
  END DO
!
!---
!
! and finally calculate H_a*H_res where H_res is : H_t*H_v*H_h
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                
  lvls_out = nlev_m/2+1
  nmax = lvls_out * 16 * Ha_csr%nrow
  
  ALLOCATE(iw(Hres2_csr%ncol))    !work array needed for routine amub
  iw=0
  CALL multiply_matrices_csr(iw,            &
                             Ha_csr,        &    !matrix A
                             Hres2_csr,     &    !matrix B
			     Hres_csr,      &    !out:A*B
			     nmax)
  CALL deallocate_operator(Ha_csr)
  DEALLOCATE(iw)
  CALL deallocate_operator(Hres2_csr)
!                                    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                
! finished calculating the H matrix,
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! 
  RETURN
END SUBROUTINE observ_operator_tc
!
!
SUBROUTINE TRANSF_TOTCOL_AVG_KERNEL(NIV,NIV_OBS,NPRF,RP,RPLT,RP_O,AVK)
!
!CCCC
!
!   VERTICAL INTEGRATION (CUBIC)
!   __  __   _____   _____   _____   _____   _____
!  |  \/  | /     \ /     \ /  _  \ /  ___\ |  ___|
!  |      | |  |  | |  <--< |  _  | |  \_ \ |  ___|
!  |_|\/|_| \_____/ \_____/ \_/ \_/ \_____/ |_____|
!
!   MODELE DE CHIMIE ATMOSPHERIQUE A GRANDE ECHELLE
!   VERSION_RELEASE_BUGFIX : 0_1_1 
!
!CC DATE : 09/06/2005
!
!   Type      : Chemistry 
!   Purpose   : Compute a total column
!   Called by : 
!
!CC
!
!  Original : B. Josse & V.-H. Peuch, 04/2000
!
!  Modified :
!
!CCCC
!	
    IMPLICIT NONE
!
!
!     Subroutine arguments.
!     ~~~~~~~~~~~~~~~~~~~~
!
    INTEGER :: NIV, NIV_OBS, NPRF
    REAL (KIND=8) , DIMENSION (NIV,NPRF) :: RP,RPLT
    REAL (KIND=8) , DIMENSION (2,NIV_OBS,NPRF) :: RP_O
    REAL (KIND=8) , DIMENSION (NIV_OBS,NIV,NPRF) :: AVK
!
! LOOPS      
    INTEGER :: I,L,K
! INDEXES
    INTEGER :: LTOP,LBOT
! COEFFICIENTS
    REAL(KIND=8) :: rl_TOP,rl_BOT
! CONSTANTS      
    REAL(KIND=8) :: MAIR,GG,AVO,DOBSON
! WEIGHTS FOR NEIGHBORING LEVELS       	
    REAL (KIND=8), ALLOCATABLE   :: ALFZ1MAS(:,:),ALFZ2MAS(:,:)
    REAL (KIND=8), ALLOCATABLE   :: ALFZ3MAS(:,:),ALFZ4MAS(:,:)
    REAL (KIND=8), ALLOCATABLE   :: ALFZ5MAS(:,:)
! WORK ARRAY
    REAL (KIND=8), DIMENSION (NIV) :: AVK_COEFF
    REAL (KIND=8), ALLOCATABLE :: RP_SWAP(:,:)
    LOGICAL :: LL_SWAP
!AP    REAL (KIND=8) :: S1
!
    ALLOCATE (ALFZ1MAS(NPRF,NIV))
    ALLOCATE (ALFZ2MAS(NPRF,NIV))
    ALLOCATE (ALFZ3MAS(NPRF,NIV))
    ALLOCATE (ALFZ4MAS(NPRF,NIV))
    ALLOCATE (ALFZ5MAS(NPRF,NIV))
!
! 1. INITIALIZATIONS
!
    DO I = 1,NPRF
      DO L=1,NIV
	DO K=1,NIV_OBS
        AVK(K,L,I)=0.
	ENDDO
      ENDDO
    ENDDO
!
!AP    s1 =sign(1.d0,RP(2,1)-RP(1,1))
    LL_SWAP = .FALSE.
    IF (RP(2,1) .LT. RP(1,1)) THEN
	ALLOCATE(RP_SWAP(NIV,NPRF))
	RP_SWAP(:,:)=RP(:,:)
	DO L=1,NIV
	  RP(L,:)=RP_SWAP(NIV-L+1,:)
	END DO
!	
	RP_SWAP(:,:)=RPLT(:,:)
	DO L=1,NIV
	  RPLT(L,:)=RP_SWAP(NIV-L+1,:)
	END DO
	LL_SWAP = .TRUE.
	DEALLOCATE(RP_SWAP)
	ALLOCATE(RP_SWAP(NIV_OBS,NIV))
    END IF
	
!
    MAIR=0.029
    GG=9.80665
    AVO=6.02E23
    DOBSON=2.685E20
!
! 2. COMPUTE WEIGHTS
!        
    CALL TRANSF_TOTCOL_WEIGHTS(NIV,NPRF,RP,RPLT &
                   ,ALFZ1MAS,ALFZ2MAS &
                   ,ALFZ3MAS,ALFZ4MAS &
                   ,ALFZ5MAS)
!
! 3. COMPUTE NUMBER OF MOLES IN EACH MODEL
!    CELL (CUBIC INTERPOLATION)
!
!!$    DO I = 1,NPRF
!!$      AVK(1,1,I)=0.75*RP(1,I)+(RPLT(1,I)-RP(1,I))&
!!$         +ALFZ2MAS(I,2)+ALFZ1MAS(I,3)
!!$      AVK(1,2,I)=ALFZ3MAS(I,2)+ALFZ2MAS(I,3)+ALFZ1MAS(I,4)
!!$      AVK(1,3,I)=ALFZ4MAS(I,2)+ALFZ3MAS(I,3)+ALFZ2MAS(I,4)+ALFZ1MAS(I,5)
!!$      DO L = 4,NIV-2
!!$        AVK(1,L,I)=ALFZ5MAS(I,L-2)+ALFZ4MAS(I,L-1)+ALFZ3MAS(I,L)&
!!$           +ALFZ2MAS(I,L+1)+ALFZ1MAS(I,L+2)
!!$      END DO
!!$      AVK(1,NIV-1,I)=ALFZ5MAS(I,NIV-3)+ALFZ4MAS(I,NIV-2)&
!!$         +ALFZ3MAS(I,NIV-1)+ALFZ2MAS(I,NIV)
!!$      AVK(1,NIV,I)=ALFZ5MAS(I,NIV-2)+ALFZ4MAS(I,NIV-1)&
!!$         +ALFZ3MAS(I,NIV)
!!$    END DO
!!$    AVK(:,:,:) = AVK(:,:,:)/GG/MAIR*AVO/DOBSON
!!$
    DO I = 1,NPRF
!AP      AVK_COEFF(1)=0.75*RP(1,I)+(RPLT(1,I)-RP(1,I))&
!AP         +ALFZ2MAS(I,2)+ALFZ1MAS(I,3)
!AP      AVK_COEFF(2)=ALFZ3MAS(I,2)+ALFZ2MAS(I,3)+ALFZ1MAS(I,4)
      AVK_COEFF(1)=ALFZ3MAS(I,1)+ALFZ2MAS(I,2)+ALFZ1MAS(I,3)
      AVK_COEFF(2)=ALFZ4MAS(I,1)+ALFZ3MAS(I,2)+ALFZ2MAS(I,3)+ALFZ1MAS(I,4)
!APEND
      AVK_COEFF(3)=ALFZ4MAS(I,2)+ALFZ3MAS(I,3)+ALFZ2MAS(I,4)+ALFZ1MAS(I,5)
      DO L = 4,NIV-2
        AVK_COEFF(L)=ALFZ5MAS(I,L-2)+ALFZ4MAS(I,L-1)+ALFZ3MAS(I,L)&
           +ALFZ2MAS(I,L+1)+ALFZ1MAS(I,L+2)
      END DO
      AVK_COEFF(NIV-1)=ALFZ5MAS(I,NIV-3)+ALFZ4MAS(I,NIV-2)&
         +ALFZ3MAS(I,NIV-1)+ALFZ2MAS(I,NIV)
      AVK_COEFF(NIV)=ALFZ5MAS(I,NIV-2)+ALFZ4MAS(I,NIV-1)&
         +ALFZ3MAS(I,NIV)
!
      AVK_COEFF(:) = AVK_COEFF(:)/GG/MAIR*AVO/DOBSON
!
! 4. DETERMINE LEVELS/COLUMNS INTERSECTIONS
!
      DO K=1,NIV_OBS ! For each subcolumn
!
	IF (ABS(RP_O(1,K,I)).lt.1.e-6 .or. RPLT(1,I).gt.RP_O(1,K,I)) THEN 
!***        Bring down the 0. or very small top boundary to the model roof
	    LTOP=1
	    rl_TOP=1.
	ELSE
	    LTOP=1
	    DO WHILE(RPLT(LTOP,I).lt.RP_O(1,K,I))
	      LTOP=LTOP+1
	    END DO
!
	    rl_TOP=(RPLT(LTOP,I)-RP_O(1,K,I))/ &
	      (RPLT(LTOP,I)-RPLT(LTOP-1,I)) 
	    
	END IF
!
	IF (RPLT(NIV,I).lt.RP_O(2,K,I)) THEN 
!***        Bring up the very large bottom boundary to the model ground
	    LBOT=NIV
	    rl_BOT=1.
	ELSE
	    LBOT=NIV
	    DO WHILE(RPLT(LBOT,I).gt.RP_O(2,K,I))
	      LBOT=LBOT-1
	    END DO
	    LBOT=LBOT+1
!
	    rl_BOT=(RP_O(2,K,I)-RPLT(LBOT-1,I))/ &
	      (RPLT(LBOT,I)-RPLT(LBOT-1,I)) 
	END IF
!
! 5. FILL AVK_KERNEL LINES
!
	AVK(K,LTOP,I)=rl_TOP*AVK_COEFF(LTOP)
	DO L = LTOP+1,LBOT-1
	  AVK(K,L,I)=AVK_COEFF(L)
	END DO
	AVK(K,LBOT,I)=rl_BOT*AVK_COEFF(LBOT)
      END DO ! K=1,NIV_OBS
      IF (LL_SWAP) THEN
	  RP_SWAP(:,:)=AVK(:,:,I)
	  DO L=1,NIV
	    AVK(:,L,I)=RP_SWAP(:,NIV-L+1)
	  END DO
      END IF
    END DO ! I = 1,NPRF 
!
! CLEAN UP
!
    IF (LL_SWAP) THEN
	DEALLOCATE(RP_SWAP)
	ALLOCATE(RP_SWAP(NIV,NPRF))
	RP_SWAP(:,:)=RP(:,:)
	DO L=1,NIV
	  RP(L,:)=RP_SWAP(NIV-L+1,:)
	END DO
!	
	RP_SWAP(:,:)=RPLT(:,:)
	DO L=1,NIV
	  RPLT(L,:)=RP_SWAP(NIV-L+1,:)
	END DO
	LL_SWAP = .TRUE.
	DEALLOCATE(RP_SWAP)
    END IF
    DEALLOCATE (ALFZ1MAS)
    DEALLOCATE (ALFZ2MAS)
    DEALLOCATE (ALFZ3MAS)
    DEALLOCATE (ALFZ4MAS)
    DEALLOCATE (ALFZ5MAS)
!    
  END SUBROUTINE TRANSF_TOTCOL_AVG_KERNEL
!
  SUBROUTINE TRANSF_TOTCOL_WEIGHTS(NIV,NPRF,PL,PLT  &
                           ,ALFZ1MAS,ALFZ2MAS  &
                           ,ALFZ3MAS,ALFZ4MAS  &
                           ,ALFZ5MAS)
!     
	IMPLICIT NONE
!	
!
!     Subroutine arguments.
!     ~~~~~~~~~~~~~~~~~~~~
!
	INTEGER :: NIV,NPRF
	REAL (KIND=8) , DIMENSION (NIV,NPRF) :: PL,PLT
	REAL (KIND=8) , DIMENSION (NPRF,NIV) :: ALFZ1MAS,ALFZ2MAS
	REAL (KIND=8) , DIMENSION (NPRF,NIV) :: ALFZ3MAS,ALFZ4MAS
	REAL (KIND=8) , DIMENSION (NPRF,NIV) :: ALFZ5MAS
!
!     Variables locales.
!     ~~~~~~~~~~~~~~~~~      
!
	INTEGER :: I,J,L,LP1
	REAL (KIND=8) :: PL1,PL2,PLT0,PLT1,PLT2,COEF,DP2
	REAL (KIND=8) :: RPN,RPNM1,RPTN,RPTNM1
	REAL (KIND=8) :: RP1,RP2,RP3,RP4,DP1
	REAL (KIND=8) :: RP12,RP13,RP14,RP23,RP24,RP34
	REAL (KIND=8) , ALLOCATABLE :: INVPROD(:,:),SOMPROD(:,:)
	REAL (KIND=8) , ALLOCATABLE :: PROD(:,:),SOM(:,:)
	REAL (KIND=8) , ALLOCATABLE :: ZCOEF(:)	
!        
	ALLOCATE (INVPROD(NIV,4))
	ALLOCATE (SOMPROD(NIV,4))
	ALLOCATE (PROD(NIV,4))
	ALLOCATE (SOM(NIV,4))
	ALLOCATE (ZCOEF(4))
!
      DO 1 I=1,NPRF
!
! 1er cas: maille superieure L=1
! Valeur prise constante dans la premiere moitie de maille,
! puis interpolation lineaire entre L=1 et L~=1
!
        PL1=PL(1,I)
        PL2=PL(2,I)
        PLT1=PLT(1,I)
        PLT0=0.
        COEF=((PLT1-PL1)**2)/(PL2-PL1)
        COEF=COEF/2.
        ALFZ1MAS(I,1)=0.
        ALFZ2MAS(I,1)=0.
        ALFZ3MAS(I,1)=(PLT1-PLT0)-COEF
        ALFZ4MAS(I,1)=COEF          
        ALFZ5MAS(I,1)=0.                    
!                  
! 2 eme cas: maille L=2.
! Interpolation lineaire dans la premiere moitie de maille
! puis cubique dans la seconde moitie (les coefficients
! beta correspondant a la seconde moitie sont calcules plus loin)
!                
        COEF=((PL2-PLT1)**2)/2.     
        COEF=COEF/(PL2-PL1)
        ALFZ1MAS(I,2)=0.
        ALFZ2MAS(I,2)=COEF
        ALFZ3MAS(I,2)=(PL2-PLT1)-COEF
        ALFZ4MAS(I,2)=0.
!
! Cas general: interpolation cubique dans les deux moities de mailles
! coefficients de la moitie superieure (courbe passant par les points
! L-2,L-1,L et L+1):ALF
!
! La masse dans la maille sera calculee par:
!  ALFZ1 X(L-2) + (ALFZ2)X(L-1) + (ALFZ3) X(L)
! + (ALFZ4) X(L+1) + ALFZ5 X(L+2)
!
! Calcul de ALF(Gauche) de L=3 a NIV-1
! Calcul de ALF(Droite) de L=2 a NIV-2
!
        DO 10 L=3,NIV-1
          RP1=PL(L-2,I)
          RP2=PL(L-1,I)
          RP3=PL(L,I)
          RP4=PL(L+1,I)
          DP1=PLT(L-1,I)
          RP12=RP1*RP2
          RP13=RP1*RP3
          RP14=RP1*RP4
          RP23=RP2*RP3
          RP24=RP2*RP4
          RP34=RP3*RP4
          INVPROD(L,1)=1./((RP1-RP2)*(RP1-RP3)*(RP1-RP4))
          INVPROD(L,2)=1./((RP2-RP1)*(RP2-RP3)*(RP2-RP4))
          INVPROD(L,3)=1./((RP3-RP1)*(RP3-RP2)*(RP3-RP4))
          INVPROD(L,4)=1./((RP4-RP1)*(RP4-RP2)*(RP4-RP3))
          SOM(L,1)=RP2+RP3+RP4
          SOM(L,2)=RP1+RP3+RP4
          SOM(L,3)=RP1+RP2+RP4
          SOM(L,4)=RP1+RP2+RP3
          PROD(L,1)=RP2*RP3*RP4
          PROD(L,2)=RP1*RP3*RP4
          PROD(L,3)=RP1*RP2*RP4
          PROD(L,4)=RP1*RP2*RP3
          SOMPROD(L,1)=RP23+RP24+RP34
          SOMPROD(L,2)=RP13+RP14+RP34
          SOMPROD(L,3)=RP12+RP14+RP24
          SOMPROD(L,4)=RP12+RP13+RP23
          DO 101 J=1,4
            ZCOEF(J)=             (RP3**4-DP1**4)/4.&
                    -SOM    (L,J)*(RP3**3-DP1**3)/3.&
                    +SOMPROD(L,J)*(RP3**2-DP1**2)/2.&
                    -PROD   (L,J)*(RP3   -DP1   )
            ZCOEF(J)=INVPROD(L,J)*ZCOEF(J)
 101      CONTINUE
          ALFZ1MAS(I,L)=ZCOEF(1)
          ALFZ2MAS(I,L)=ZCOEF(2)
          ALFZ3MAS(I,L)=ZCOEF(3)
          ALFZ4MAS(I,L)=ZCOEF(4)
 10     CONTINUE
!
        DO 11 L=2,NIV-2
           LP1=L+1
           RP3=PL(L,I)
           RP4=PL(L+1,I)
           DP2=PLT(L,I)
           DO 111 J=1,4
             ZCOEF(J)=               (DP2**4-RP3**4)/4.&
                     -SOM    (LP1,J)*(DP2**3-RP3**3)/3.&
                     +SOMPROD(LP1,J)*(DP2**2-RP3**2)/2.&
                     -PROD   (LP1,J)*(DP2   -RP3   )
             ZCOEF(J)=INVPROD(LP1,J)*ZCOEF(J)
 111       CONTINUE
           ALFZ2MAS(I,L)=ALFZ2MAS(I,L)+ZCOEF(1)
           ALFZ3MAS(I,L)=ALFZ3MAS(I,L)+ZCOEF(2)
           ALFZ4MAS(I,L)=ALFZ4MAS(I,L)+ZCOEF(3)
           ALFZ5MAS(I,L)=ZCOEF(4)
 11     CONTINUE
!
! Avant-derniere maille (pres du sol: L=NIV-1)
! Meme principe que pour L=2
! Les ALFZ ont ete calcules plus haut
!
        RPN   =PL (NIV,I)
        RPNM1 =PL (NIV-1,I)
        RPTN  =PLT(NIV,I)
        RPTNM1=PLT(NIV-1,I)
        COEF=(RPTNM1**2-RPNM1**2)/2.-RPN*(RPTNM1-RPNM1)
        COEF=COEF/(RPNM1-RPN)         
        ALFZ3MAS(I,NIV-1)=ALFZ3MAS(I,NIV-1)+               COEF
        ALFZ4MAS(I,NIV-1)=ALFZ4MAS(I,NIV-1)+(RPTNM1-RPNM1)-COEF
        ALFZ5MAS(I,NIV-1)=0.
!
! Derniere maille (L=NIV)
! meme principe que pour L=1
!
        COEF=(RPN**2-RPTNM1**2)/2.-RPN*(RPN-RPTNM1)
        COEF=COEF/(RPNM1-RPN)
!
! essai d'interpolation lineaire jusqu'au sol
! sans correction en cas de passage sous 0
!
        ALFZ1MAS(I,NIV)=0.
        ALFZ2MAS(I,NIV)=           -COEF
        ALFZ3MAS(I,NIV)=RPTN-RPTNM1+COEF
        ALFZ4MAS(I,NIV)=0.         
        ALFZ5MAS(I,NIV)=0.
!         
 1    CONTINUE     
!
	DEALLOCATE (INVPROD)
	DEALLOCATE (SOMPROD)
	DEALLOCATE (PROD)
	DEALLOCATE (SOM)
	DEALLOCATE (ZCOEF)
!
  END SUBROUTINE TRANSF_TOTCOL_WEIGHTS
