      INTEGER FUNCTION DMESEC2(KGRIB,KLENG,INSPT,KSEC2,IBITS,ILEN2)
C
C---->
C**** DMESEC2P
C
C     Purpose
C     -------
C
C     Decodes GRIB section 2 values for mercator fields.
C
C     Interface
C     ---------
C
C     IRET = DMESEC2(KGRIB,KLENG,INSPT,KSEC2,IBITS)
C
C     Input parameters
C     ----------------
C
C     KGRIB   - GRIB product unpacked so far (upto octet 6). 
C     KLENG   - Length of KGRIB.
C     INSPT   - Bit position of GRIB product unpacked so far.
C     KSEC2   - GRIB section 2 description.
C     IBITS   - Number of bits per element in KGRIB.
C     ILEN2   - Length in bytes of section 2
C     
C
C     Output parameters
C     -----------------
C
C     KSEC2   - Updated GRIB section 2 description.
C     INSPT   - Updated bit position of GRIB product built so far.
C     IRET    - Function return status code, 0 = OK
C
C     Common block usage
C     ------------------
C
C     None.
C
C     Method
C     ------
C
C     Unpacks values given in KGRIB into KSEC2 and updates bit
C     pointer INSPT.
C
C     Externals
C     ---------
C
C     DSGNBT  - Get the sign bit and adjust value to +/-.
C     INXBIT  - Insert bit data.
C
C
C     Author
C     ------
C
C     J.D.Chambers       ECMWF July 2004.
C
C     Modifications
C     -------------
C
C     None.
C
C----<
C     ----------------------------------------------------------------|
C     Section 0. Definition of variables.
C     ----------------------------------------------------------------|
C
      IMPLICIT NONE
C
#include "grprs.h"
C
C     Parameters
C
      INTEGER JP16SET
      PARAMETER ( JP16SET = 2**16 - 1 ) 
C                            ^---> 65535    =   FFFF(hex)
C
C     Function arguments
C
      INTEGER KGRIB,KLENG,INSPT,KSEC2,IBITS,ILEN2
      DIMENSION KGRIB(*),KSEC2(*)
C
C     Local variables
C
      INTEGER ILALO(2), IRET, IRESOL
C
C     ----------------------------------------------------------------|
C     Section 1.    Initialization.
C     ----------------------------------------------------------------|
C
  100 CONTINUE
C
      DMESEC2 = 0
C
C     ----------------------------------------------------------------|
C     Section 2.    Unpack section 2 octets.
C     ----------------------------------------------------------------|
C
  200 CONTINUE
C
C      Octets 7 - 8 : Nx - number of points along a parallel.
C      Octets 9 - 10: Ny - number of points along a meridian.
C      Two 16 bit fields.
C 
       CALL INXBIT(KGRIB,KLENG,INSPT,KSEC2(2),2,IBITS, 16,'D',IRET)
       IF( IRET.NE.0 ) THEN
         DMESEC2 = 1
         WRITE(GRPRSM,*) 'DMESEC2: Error extracting number of points'
         WRITE(GRPRSM,*) 'DMESEC2: along parallel or meridian.'
         WRITE(GRPRSM,*) 'DMESEC2: Return code = ', IRET
         GO TO 900
       ENDIF
C 
C      Octets 11 - 13: La1 - latitude of first grid-point.
C      Octets 14 - 16: Lo1 - longitude of first grid-point.
C      Two 24 bit fields.
C 
       CALL INXBIT(KGRIB,KLENG,INSPT,ILALO(1),2,IBITS, 24,'D',IRET)
       IF( IRET.NE.0 ) THEN
         DMESEC2 = 1
         WRITE(GRPRSM,*) 'DMESEC2: Error extracting'
         WRITE(GRPRSM,*)
     X     'DMESEC2: latitude/longitude of first grid-point.'
         WRITE(GRPRSM,*) 'DMESEC2: Return code = ', IRET
         GO TO 900
       ENDIF
C 
C      If sign bit is 1, value is negative.
C 
       CALL DSGNBT( KSEC2(4), ILALO(1), 24, IRET)
       CALL DSGNBT( KSEC2(5), ILALO(2), 24, IRET)
C 
C      Octet 17: Resolution and components flag.
C      One 8 bit field.
C 
       CALL INXBIT(KGRIB,KLENG,INSPT,IRESOL,1,IBITS, 8,'D',IRET)
       IF( IRET.NE.0 ) THEN
         DMESEC2 = 1
         WRITE(GRPRSM,*) 'DMESEC2: Error extracting'
         WRITE(GRPRSM,*) 'DMESEC2: resolution and components flag.'
         WRITE(GRPRSM,*) 'DMESEC2: Return code = ', IRET
         GO TO 900
       ENDIF
C 
C      All flag fields are already set to 0, so
C
       IF( IRESOL.EQ.0 ) GO TO 214
C
C      Resolution flag is not applicable.
C
       IF( IRESOL.GE.128) IRESOL = IRESOL - 128
C
C      Set earth flag.
C
       IF( IRESOL.GE.64) THEN
         KSEC2(18) = 64
         IRESOL    = IRESOL - 64
       ENDIF
C 
C      Set components flag.
C 
       KSEC2(19) = IRESOL
C 
  214  CONTINUE
C
C      Octets 18 - 20: La1 - latitude of last grid-point.
C      Octets 21 - 23: Lo1 - longitude of last grid-point.
C      Two 24 bit fields.
C 
       CALL INXBIT(KGRIB,KLENG,INSPT,ILALO(1),2,IBITS, 24,'D',IRET)
       IF( IRET.NE.0 ) THEN
         DMESEC2 = 1
         WRITE(GRPRSM,*) 'DMESEC2: Error extracting'
         WRITE(GRPRSM,*)
     X     'DMESEC2: latitude/longitude of last grid-point.'
         WRITE(GRPRSM,*) 'DMESEC2: Return code = ', IRET
         GO TO 900
       ENDIF
C 
C      If sign bit is 1, value is negative.
C 
       CALL DSGNBT( KSEC2(7), ILALO(1), 24, IRET)
       CALL DSGNBT( KSEC2(8), ILALO(2), 24, IRET)
C     
C      Octets 24 - 26: Latin - latitude at which the mercator
C                               projection cylinder intersects the Earth
C      One 24 bit field.
C
       CALL INXBIT(KGRIB,KLENG,INSPT,ILALO(1),1,IBITS, 24,'D',IRET)
       IF( IRET.NE.0 ) THEN
         DMESEC2 = 1
         WRITE(GRPRSM,*) 'DMESEC2: Error extracting Latin'
         WRITE(GRPRSM,*) 'DMESEC2: Return code = ', IRET
         GO TO 900
       ENDIF
C     
C      If sign bit is 1, value is negative.
C
       CALL DSGNBT( KSEC2(9), ILALO(1), 24, IRET)
C 
C      Octet 27: Reserved (set to zero)
C      Octet 28: Scanning mode flags.
C      Two 8 bit fields.
C 
       INSPT = INSPT + 8
       KSEC2(10) = 0
       CALL INXBIT(KGRIB,KLENG,INSPT,KSEC2(11),1,IBITS, 8,'D',IRET)
       IF( IRET.NE.0 ) THEN
         DMESEC2 = 1
         WRITE(GRPRSM,*) 'DMESEC2: Error extracting'
         WRITE(GRPRSM,*) 'DMESEC2: scanning mode flags.'
         WRITE(GRPRSM,*) 'DMESEC2: Return code = ', IRET
         GO TO 900
       ENDIF
C 
C     Octets 29 - 31: Di - longitudinal direction grid length
C     Octets 32 - 34: Di - latitudinal direction grid length
C     Two 24 bit fields.
C 
      CALL INXBIT(KGRIB,KLENG,INSPT,KSEC2(13),2,IBITS, 24,'D',IRET)
      IF( IRET.NE.0 ) THEN
        DMESEC2 = 1
        WRITE(GRPRSM,*) 'DMESEC2: Error extracting'
        WRITE(GRPRSM,*)
     X     'DMESEC2: longitudinal/latitudinal direction grid lengths'
        WRITE(GRPRSM,*) 'DMESEC2: Return code = ', IRET
        GO TO 900
      ENDIF
C 
C     Octets 35 - 42: Reserved (set to zero)
C     Eight 8 bit fields.
C
      INSPT = INSPT + (ILEN2-34)*8
C
      KSEC2(17) = 0
      KSEC2(20) = 0
      KSEC2(21) = 0
      KSEC2(22) = 0
C 
C     ----------------------------------------------------------------|
C     Section 9. Return to calling routine.
C     ----------------------------------------------------------------|
C 
  900 CONTINUE
      RETURN
      END
