      INTEGER FUNCTION EMESEC2(KGRIB,KLENG,INSPT,KSEC2,IBITS,ILEN2)
C
C---->
C**** EMESEC2
C
C     Purpose
C     -------
C
C     Encodes GRIB section 2 values for mercator fields.
C
C     Interface
C     ---------
C
C     IRET = EMESEC2(KGRIB,KLENG,INSPT,KSEC2,IBITS)
C
C     Input parameters
C     ----------------
C
C     KGRIB   - GRIB product built so far (upto octet 6). 
C     KLENG   - Length of KGRIB.
C     INSPT   - Bit position of GRIB product built so far.
C     KSEC2   - GRIB section 2 description.
C     IBITS   - Number of bits per element in KGRIB.
C     ILEN2   - Length in bytes of section 2
C     
C
C     Output parameters
C     -----------------
C
C     KGRIB   - Updated GRIB product to end of section 2.
C     INSPT   - Updated bit position of GRIB product built so far.
C     IRET    - Function return status code, 0 = OK
C
C     Common block usage
C     ------------------
C
C     None.
C
C     Method
C     ------
C
C     Packs values given in KSEC2 into KGRIB and updates bit
C     pointer INSPT.
C
C     Externals
C     ---------
C
C     CSGNBT  - Set the sign bit.
C     INXBIT  - Insert bit data.
C
C
C     Author
C     ------
C
C     J.D.Chambers       ECMWF July 2004.
C
C     Modifications
C     -------------
C
C     None.
C
C----<
C     ----------------------------------------------------------------|
C     Section 0. Definition of variables.
C     ----------------------------------------------------------------|
C
      IMPLICIT NONE
C
#include "grprs.h"
C
C     Parameters
C
C
C     Function arguments
C
      INTEGER KGRIB,KLENG,INSPT,KSEC2,IBITS,ILEN2
      DIMENSION KGRIB(*),KSEC2(*)
C
C     Local variables
C
      INTEGER ILALO(2), IRET, IRESOL, IZERO, LOOP
      DATA IZERO/0/
C
C     ----------------------------------------------------------------|
C     Section 1.    Initialization.
C     ----------------------------------------------------------------|
C
  100 CONTINUE
C
      EMESEC2 = 0
C
C     ----------------------------------------------------------------|
C     Section 2.    Pack section 2 octets.
C     ----------------------------------------------------------------|
C
  200 CONTINUE
C
C     Octets 7 - 8 : Nx - number of points along a parallel.
C     Octets 9 - 10: Ny - number of points along a meridian.
C     Two 16 bit fields.
C
      CALL INXBIT(KGRIB,KLENG,INSPT,KSEC2(2),2,IBITS, 16,'C',IRET)
      IF( IRET.NE.0 ) THEN
        EMESEC2 = 1
        WRITE(GRPRSM,*) 'EMESEC2: Error inserting number of points'
        WRITE(GRPRSM,*) 'EMESEC2: along parallel or meridian.'
        WRITE(GRPRSM,*) 'EMESEC2: Return code = ', IRET
        GOTO 900
      ENDIF
C
C     Octets 11 - 13: Lap - latitude of first grid-point.
C     Octets 14 - 16: Lop - longitude of first grid-point.
C     Two 24 bit fields.
C
C     Set sign bit to 1, if value is negative.
C
      CALL CSGNBT( ILALO(1), KSEC2(4), 24, IRET)
      CALL CSGNBT( ILALO(2), KSEC2(5), 24, IRET)
C
      CALL INXBIT(KGRIB,KLENG,INSPT,ILALO(1),2,IBITS, 24,'C',IRET)
      IF( IRET.NE.0 ) THEN
        EMESEC2 = 1
        WRITE(GRPRSM,*) 'EMESEC2: Error inserting'
        WRITE(GRPRSM,*)
     X    'EMESEC2: latitude/longitude of first grid-point.'
        WRITE(GRPRSM,*) 'EMESEC2: Return code = ', IRET
        GOTO 900
      ENDIF
C
C     Octet 17: Resolution and components flag.
C     One 8 bit field.
C
C     Resolution flag ( KSEC2(6) ) is not applicable.
C
      IRESOL = KSEC2(18)+KSEC2(19)
C
      CALL INXBIT(KGRIB,KLENG,INSPT,IRESOL,1,IBITS, 8,'C',IRET)
      IF( IRET.NE.0 ) THEN
        EMESEC2 = 1
        WRITE(GRPRSM,*) 'EMESEC2: Error inserting'
        WRITE(GRPRSM,*) 'EMESEC2: components flag.'
        WRITE(GRPRSM,*) 'EMESEC2: Return code = ', IRET
        GOTO 900
      ENDIF
C
C     Octets 18 - 20: La1 - latitude of last grid-point.
C     Octets 21 - 23: Lo1 - longitude of last grid-point.
C     Two 24 bit fields.
C
C     Set sign bit to 1, if value is negative.
C
      CALL CSGNBT( ILALO(1), KSEC2(7), 24, IRET)
      CALL CSGNBT( ILALO(2), KSEC2(8), 24, IRET)

      CALL INXBIT(KGRIB,KLENG,INSPT,ILALO(1),2,IBITS, 24,'C',IRET)
      IF( IRET.NE.0 ) THEN
        EMESEC2 = 1
        WRITE(GRPRSM,*) 'EMESEC2: Error inserting'
        WRITE(GRPRSM,*) 'EMESEC2: latitude/longitude of last grid-point'
        WRITE(GRPRSM,*) 'EMESEC2: Return code = ', IRET
        GOTO 900
      ENDIF
C     
C     Octets 24 - 26: Latin - latitude at which the mercator
C                             projection cylinder intersects the Earth
C     One 24 bit field.
C
C     Set sign bit to 1, if value is negative.
C
      CALL CSGNBT( ILALO(1), KSEC2(9), 24, IRET)

      CALL INXBIT(KGRIB,KLENG,INSPT,ILALO(1),1,IBITS, 24,'C',IRET)
      IF( IRET.NE.0 ) THEN 
        EMESEC2 = 1
        WRITE(GRPRSM,*) 'EMESEC2: Error inserting Latin'
        WRITE(GRPRSM,*) 'EMESEC2: Return code = ', IRET
        GOTO 900
      ENDIF
C
C     Octet 27: Reserved (set to zero)
C     One 8 bit field.
C
      CALL INXBIT(KGRIB,KLENG,INSPT,IZERO,1,IBITS, 8,'C',IRET)
      IF( IRET.NE.0 ) THEN
        EMESEC2 = 1
        WRITE(GRPRSM,*) 'EMESEC2: Error inserting'
        WRITE(GRPRSM,*) 'EMESEC2: zero'
        WRITE(GRPRSM,*) 'EMESEC2: Return code = ', IRET
        GOTO 900
      ENDIF
C
C     Octet 28: Scanning mode flags.
C     One 8 bit field.
C
      CALL INXBIT(KGRIB,KLENG,INSPT,KSEC2(11),1,IBITS, 8,'C',IRET)
      IF( IRET.NE.0 ) THEN
        EMESEC2 = 1
        WRITE(GRPRSM,*) 'EMESEC2: Error inserting'
        WRITE(GRPRSM,*) 'EMESEC2: scanning mode flags.'
        WRITE(GRPRSM,*) 'EMESEC2: Return code = ', IRET
        GOTO 900
      ENDIF
C
C     Octets 29 - 31: Di - longitudinal direction grid length
C     Octets 32 - 34: Di - latitudinal direction grid length
C     Two 24 bit fields.
C
      CALL INXBIT(KGRIB,KLENG,INSPT,KSEC2(13),2,IBITS, 24,'C',IRET)
      IF( IRET.NE.0 ) THEN
        EMESEC2 = 1
        WRITE(GRPRSM,*) 'EMESEC2: Error inserting'
        WRITE(GRPRSM,*)
     X    'EMESEC2: longitudinal/latitudinal direction grid lengths'
        WRITE(GRPRSM,*) 'EMESEC2: Return code = ', IRET
        GOTO 900
      ENDIF
C
C     Octets 35 - 42: Reserved (set to zero)
C     Eight 8 bit fields.
C
C     Fill reserved octets with 0.
C
      DO LOOP = 1, 8
        CALL INXBIT(KGRIB,KLENG,INSPT, IZERO, 1,IBITS, 8,'C',IRET)
        IF( IRET.NE.0 ) THEN
          IRET = 810
          WRITE(GRPRSM,*) 'EMESEC2: Error inserting dummy zero.'
          WRITE(GRPRSM,*) 'EMESEC2: Return code = ', IRET
          GOTO 900
        ENDIF
      ENDDO
C
C     ----------------------------------------------------------------|
C     Section 9. Return to calling routine.
C     ----------------------------------------------------------------|
C
  900 CONTINUE
      RETURN
      END
