MODULE CONTROL_VECTORS

!   Author.
!   -------
!     Y. Tremolet
!
!   Modifications.
!   -------------- 
!     Original   02-08-01  
!     Y.Tremolet 02-08-20  Split in several files.
!     C.Fischer  03-03-05  Aladin spectral treatment added.
!      N.B.1: the mean wind profiles, 2*nflev values, resp. mean U and mean V,
!             are treated as an extra appended stripe to the control vector
!             (addressed between ibgn_la and iend_la, as proposed by yannick).
!             however, they do actually belong to the Aladin state vector as well,
!             so that I decided to include them inside the spectral_fields
!             f90 structure (spec%sp1d). this somewhat mixed situation looks
!             uggly, but appears to me as the best compromise between 2 conflicting
!             properties of these bloody dangerous coefficients: they are part
!             of the spectral state, but they are not at all distributed the
!             same way as the usual (horizontal) spectral fields.
!             Actually, the mean wind profiles are assumed to be duplicated on
!             all processors (the Aladin/IFS code should make sure that SPA1
!             is indeed present and equal on all procs, as output from the direct
!             spectral transforms).
!      N.B.2: the 1D spectral profiles in the control vector are controlled
!             by two distinct variables:
!             * key yd_cva%lam1d  tells you that this appendix indeed should be
!               present in the control
!             * integer yd_cva%ns1d tells you how many fields it contains. This is
!               usually 2 (or 0 if absent) for zonal and meridional mean winds
!             Note that the CONTROL_VECTORS module now handles two distinct
!             logicals for aladin: one simply tells that we must handle the
!             Aladin horizontal spectra (key LELAM in CONTROL_VECTOR_DATA) while
!             the other tells that the mean wind is or not in the cv (key yd_cva%lam1d
!             which used to be yd_cva%lelam in Yannick's original version).
!      N.B.3: for Ald, very strictly coding, one should also modify the
!             nserr truncation into nserr+nmserr to level up with nsmax+nmsmax.
!             the present use of only one nserr dictates that the model error
!             control variable only works for square domains.
!      N.B.4: the mean wind component would also be missing in the model error
!             term since it belongs to the state vector.
!     M.Fisher 2003-11-27  Wavelet Jb and coding-norm compliance
!        Y.Tremolet    18-Mar-2004 Add CTLVEC_NORM
!        Y.Tremolet    23-Mar-2004 Made initial condition optional
!        C.Soci 16-Oct-2004  initialize nmsmax in setup_ctlvec
! ------------------------------------------------------------------

USE PARKIND1  ,ONLY : JPIM     ,JPRB
USE CONTROL_VECTORS_DATA, ONLY: nsmax, nmsmax, nserr, nobtov, nobtovg, nvatovv, nparam, &
                              & nflevl, nflevg, mytovoff, nctl3, nctl2, nctl1, &
                              & ngrb, mylevs, linitcv, ltovscv, lmoderr, lelam, &
                              & lwaveletjb, nproma, nscales, nblocksl, nblocksg, &
                              & ngptotl, ngptotg, nprgpns, nprgpew, myseta, mysetb
USE SPECTRAL_FIELDS, ONLY : SPECTRAL_FIELD, ASSIGNMENT(=), INQ_SPEC, ALLOCATE_SPEC, &
                          & DEALLOCATE_SPEC
USE MPL_MODULE, ONLY : MPL_ALLREDUCE

IMPLICIT NONE
PRIVATE
PUBLIC control_vector, setup_ctlvec, allocate_ctlvec, deallocate_ctlvec, &
     & ASSIGNMENT(=), OPERATOR(.EQV.), OPERATOR(.NEQV.), INVERSE, &
     & DOT_PRODUCT, MAXVAL, MINVAL, SUM, CTLVEC_SQNORM, CTLVEC_NORM

! ------------------------------------------------------------------

TYPE CONTROL_VECTOR
  REAL(KIND=JPRB), POINTER :: data(:) => NULL()
  LOGICAL :: lalloc = .false.
  INTEGER(KIND=JPIM) :: nsizel, nsizeg, nscales
  LOGICAL :: linitcv, lmoderr, ltovscv, lam1d, lwaveletjb
  INTEGER(KIND=JPIM) :: nsmax, nmsmax, nserr, ns3d, ns2d, ns1d, npara
  INTEGER(KIND=JPIM), POINTER :: ibgn_scale(:) => NULL()
  REAL(KIND=JPRB), POINTER :: initcv(:) => NULL()
  REAL(KIND=JPRB), POINTER :: moderr(:) => NULL()
  REAL(KIND=JPRB), POINTER :: tovscv(:) => NULL()
  REAL(KIND=JPRB), POINTER :: lamcv (:) => NULL()
  REAL(KIND=JPRB), POINTER :: params(:) => NULL()
END TYPE CONTROL_VECTOR

! ------------------------------------------------------------------

INTERFACE allocate_ctlvec
MODULE PROCEDURE allocate_cv, allocate_cv_cv
END INTERFACE

INTERFACE ASSIGNMENT (=)
MODULE PROCEDURE assign_cv_cv, assign_scalar_cv, assign_ar_cv, assign_cv_ar
END INTERFACE

INTERFACE INVERSE
MODULE PROCEDURE inverse_cv, inverse_cv_cv
END INTERFACE

INTERFACE DOT_PRODUCT
MODULE PROCEDURE dot_product_cv_cv, dot_product_weight
END INTERFACE

INTERFACE CTLVEC_SQNORM
MODULE PROCEDURE cv_sqnorm1, cv_sqnorm4
END INTERFACE

INTERFACE MAXVAL
MODULE PROCEDURE maxval_cv
END INTERFACE

INTERFACE MINVAL
MODULE PROCEDURE minval_cv
END INTERFACE

INTERFACE SUM
MODULE PROCEDURE sum_cv
END INTERFACE

INTERFACE OPERATOR (.EQV.)
MODULE PROCEDURE equiv_cv
END INTERFACE

INTERFACE OPERATOR (.NEQV.)
MODULE PROCEDURE nequiv_cv
END INTERFACE

INTERFACE check
MODULE PROCEDURE check_cv, check_cv_cv
END INTERFACE

! ------------------------------------------------------------------
CONTAINS
! ------------------------------------------------------------------

SUBROUTINE setup_ctlvec(kctl3,kctl2,kctl1,kgrb,ksmax,kmsmax,ldmerr,kserr, &
                      & ldtovs,kobtov,kobtovg,kvatovv,ldlam,kflevl,kflevg, &
                      & kvatovoff,kylevs,kparam,ldwavelet,kblocksl,kblocksg, &
                      & knscales,knproma,kgptotl,kgptotg,kprgpns,kprgpew, &
                      & kseta,ksetb,ldinit)

IMPLICIT NONE

INTEGER(KIND=JPIM), INTENT(IN) :: kctl3,kctl2,kctl1,ksmax,kmsmax,kserr,kparam
INTEGER(KIND=JPIM), INTENT(IN) :: kgrb(kctl3+kctl2)
LOGICAL, INTENT(IN) :: ldinit, ldmerr, ldtovs, ldwavelet
INTEGER(KIND=JPIM), INTENT(IN) :: kobtov, kobtovg, kvatovv
LOGICAL, INTENT(IN) :: ldlam
INTEGER(KIND=JPIM), INTENT(IN) :: kflevl, kflevg, kvatovoff, knscales, knproma, kprgpns, kprgpew, &
                       & kseta, ksetb
INTEGER(KIND=JPIM), INTENT(IN) :: kylevs(kflevl), kblocksl(knscales), kblocksg(knscales), &
                       & kgptotl(knscales,kprgpns,kprgpew), kgptotg(knscales)

nctl2=kctl2
nctl3=kctl3
IF (.NOT.ALLOCATED(ngrb)) ALLOCATE(ngrb(nctl3+nctl2))
ngrb(:)=kgrb(:)
nctl1=kctl1
nsmax=ksmax
nmsmax=kmsmax
lmoderr=ldmerr
linitcv=ldinit
nserr=kserr
ltovscv=ldtovs
nobtov =kobtov
nobtovg=kobtovg
nvatovv=kvatovv
lelam=ldlam
nflevl=kflevl
nflevg=kflevg
nproma=knproma
mytovoff=kvatovoff
nprgpns=kprgpns
nprgpew=kprgpew
myseta=kseta
mysetb=ksetb
IF (.NOT.ALLOCATED(MYLEVS)) ALLOCATE(MYLEVS(nflevl))
MYLEVS(:)=kylevs(:)
nparam=kparam
lwaveletjb=ldwavelet
IF (ldwavelet) THEN
  nscales=knscales
  ALLOCATE (nblocksl(knscales))
  ALLOCATE (nblocksg(knscales))
  ALLOCATE (ngptotl(knscales,kprgpns,kprgpew))
  ALLOCATE (ngptotg(knscales))
  nblocksl(:)=kblocksl(:)
  nblocksg(:)=kblocksg(:)
  ngptotl(:,:,:)=kgptotl(:,:,:)
  ngptotg(:)=kgptotg(:)
ENDIF

write(*,*)'SETUP_CTLVEC: CONTROL_VECTOR type has been setup.'
write(*,*)'SETUP_CTLVEC: LINITCV, LMODERR, LTOVSCV, LELAM, NPARAM, LWAVELETJB=', &
                       & LINITCV, LMODERR, LTOVSCV, LELAM, NPARAM, LWAVELETJB

END SUBROUTINE setup_ctlvec

! ------------------------------------------------------------------

SUBROUTINE allocate_cv(yd_cv,ksmax,kmsmax,kserr,kparam,ldinit,ldmerr,ldtovs,ldlam,ks3d,ks2d,ks1d,knscales)

TYPE(control_vector), INTENT(OUT) :: yd_cv
INTEGER(KIND=JPIM), OPTIONAL, INTENT(IN) :: ksmax,kmsmax,kserr,kparam
LOGICAL, OPTIONAL, INTENT(IN) :: ldinit,ldmerr,ldtovs,ldlam
INTEGER(KIND=JPIM), OPTIONAL, INTENT(IN) :: ks3d, ks2d, ks1d, knscales

INTEGER(KIND=JPIM) :: ilen,ibgn_ic,iend_ic,ibgn_me,iend_me,ibgn_tv,iend_tv,ibgn_la,iend_la, &
                  &   ibgn_pa,iend_pa,j,ibgn
TYPE(SPECTRAL_FIELD) :: yl_zfld

! Default values
yd_cv%nsmax = nsmax
yd_cv%nmsmax = nmsmax
yd_cv%nserr = nserr
yd_cv%linitcv=linitcv
yd_cv%lmoderr=lmoderr
yd_cv%ltovscv=ltovscv
yd_cv%lam1d  =lelam
yd_cv%ns2d  = nctl2
yd_cv%ns3d  = nctl3
yd_cv%ns1d  = nctl1
yd_cv%npara = nparam
yd_cv%lwaveletjb = lwaveletjb
if (lwaveletjb) then
  yd_cv%nscales = nscales
endif

! Check for optional arguments
IF (PRESENT(ksmax)) yd_cv%nsmax=ksmax
IF (PRESENT(kmsmax)) yd_cv%nmsmax=kmsmax
IF (PRESENT(kserr)) yd_cv%nserr=kserr
IF (PRESENT(ldmerr)) yd_cv%lmoderr=ldmerr
IF (PRESENT(ldinit)) yd_cv%linitcv=ldinit
IF (PRESENT(ldtovs)) yd_cv%ltovscv=ldtovs
IF (PRESENT(ldlam )) yd_cv%lam1d  =ldlam
IF (PRESENT(ks2d)) yd_cv%ns2d=ks2d
IF (PRESENT(ks3d)) yd_cv%ns3d=ks3d
IF (PRESENT(ks1d)) yd_cv%ns1d=ks1d
IF (PRESENT(kparam)) yd_cv%npara=kparam
IF (PRESENT(knscales)) yd_cv%nscales=knscales

yd_cv%lwaveletjb = yd_cv%lwaveletjb .AND. yd_cv%linitcv
IF (PRESENT(knscales)) THEN
  IF (knscales >NSCALES) THEN
    write (*,*) 'knscales too large: knscales=',knscales,' > NSCALES=',NSCALES
    CALL ABOR1 ('knscales>NSCALES in allocate_cv')
  ENDIF
ENDIF

!cf To avoid stupid array bound violation
IF (yd_cv%ns1d==0) yd_cv%lam1d=.false.
yd_cv%nsizel=0
yd_cv%nsizeg=0

IF (yd_cv%linitcv) THEN
  if (.not.yd_cv%lwaveletjb) then
    CALL inq_spec(yl_zfld,yd_cv%nsmax,yd_cv%nmsmax,yd_cv%ns3d,yd_cv%ns2d,yd_cv%ns1d)
    ilen=yl_zfld%nspsizel
    ibgn_ic=1
    iend_ic=ilen
    yd_cv%nsizel=ilen
    ilen=yl_zfld%nspsizeg
    yd_cv%nsizeg=ilen
    CALL deallocate_spec(yl_zfld)
  else
    if (yd_cv%nscales==NSCALES) then
      ilen=nproma*(nflevg*yd_cv%ns3d+yd_cv%ns2d)*sum(nblocksl(:))
    else
      ilen = nproma*(nflevg*yd_cv%ns3d+yd_cv%ns2d) &
          & *(2*nblocksl(2+NSCALES-yd_cv%nscales)+sum(nblocksl(3+NSCALES-yd_cv%nscales:)))
    endif
    ibgn_ic=1
    iend_ic=ilen
    yd_cv%nsizel=ilen
    if (yd_cv%nscales==NSCALES) then
      yd_cv%nsizeg=(nflevg*yd_cv%ns3d+yd_cv%ns2d)*sum(ngptotg(:))
    else
      yd_cv%nsizeg = (nflevg*yd_cv%ns3d+yd_cv%ns2d) &
               & *(2*ngptotg(2+NSCALES-yd_cv%nscales)+sum(ngptotg(3+NSCALES-yd_cv%nscales:)))
    endif
  endif
ENDIF

IF (yd_cv%lmoderr) THEN
  CALL inq_spec(yl_zfld,yd_cv%nserr,yd_cv%nserr,yd_cv%ns3d,yd_cv%ns2d,yd_cv%ns1d)
  ilen=yl_zfld%nspsizel
  ibgn_me = yd_cv%nsizel + 1
  iend_me = yd_cv%nsizel + ilen
  yd_cv%nsizel = yd_cv%nsizel + ilen
  ilen=yl_zfld%nspsizeg
  yd_cv%nsizeg = yd_cv%nsizeg + ilen
  CALL deallocate_spec(yl_zfld)
ENDIF

IF (yd_cv%ltovscv) THEN
  ilen = nobtov*nvatovv
  ibgn_tv = yd_cv%nsizel + 1
  iend_tv = yd_cv%nsizel + ilen
  yd_cv%nsizel = yd_cv%nsizel + ilen
  yd_cv%nsizeg = yd_cv%nsizeg + nobtovg*nvatovv
ENDIF

IF (yd_cv%lam1d) THEN
  ilen=nflevl*yd_cv%ns1d
  ibgn_la = yd_cv%nsizel + 1
  iend_la = yd_cv%nsizel + ilen
  yd_cv%nsizel = yd_cv%nsizel + ilen
  yd_cv%nsizeg = yd_cv%nsizeg + nflevg*yd_cv%ns1d
ENDIF

IF (yd_cv%npara>0) THEN
  ibgn_pa = yd_cv%nsizel + 1
  iend_pa = yd_cv%nsizel + yd_cv%npara
  yd_cv%nsizel = yd_cv%nsizel + yd_cv%npara
  yd_cv%nsizeg = yd_cv%nsizeg + yd_cv%npara
ENDIF

ALLOCATE(yd_cv%data(1:yd_cv%nsizel))
yd_cv%lalloc = .true.

IF (yd_cv%linitcv) THEN
  yd_cv%initcv => yd_cv%data(ibgn_ic:iend_ic)

  IF (yd_cv%lwaveletjb) THEN
    ALLOCATE(yd_cv%ibgn_scale(yd_cv%nscales))
    ibgn=ibgn_ic
    DO j=1,yd_cv%nscales
      yd_cv%ibgn_scale(j) = ibgn
      if (yd_cv%nscales==NSCALES) then
        ilen = nproma*(nflevg*yd_cv%ns3d+yd_cv%ns2d)*nblocksl(j)
      else
        ilen = nproma*(nflevg*yd_cv%ns3d+yd_cv%ns2d)*nblocksl(max(2,j)+NSCALES-yd_cv%nscales)
      endif
      ibgn = ibgn + ilen
    ENDDO
    if (ibgn /= 1+iend_ic) CALL ABOR1 ('Error defining indices for initcv')
  ENDIF
ENDIF

IF (yd_cv%lmoderr) yd_cv%moderr => yd_cv%data(ibgn_me:iend_me)
IF (yd_cv%ltovscv) yd_cv%tovscv => yd_cv%data(ibgn_tv:iend_tv)
IF (yd_cv%lam1d  ) yd_cv%lamcv  => yd_cv%data(ibgn_la:iend_la)
IF (yd_cv%npara>0) yd_cv%params => yd_cv%data(ibgn_pa:iend_pa)

END SUBROUTINE allocate_cv

! ------------------------------------------------------------------

SUBROUTINE allocate_cv_cv(yd_cv,yd_cvin)
TYPE(control_vector), INTENT(OUT) :: yd_cv
TYPE(control_vector), INTENT(IN) :: yd_cvin
LOGICAL :: llinit,llmerr,lltovs,lllam

llinit=yd_cvin%linitcv
llmerr=yd_cvin%lmoderr
lltovs=yd_cvin%ltovscv
lllam =yd_cvin%lam1d

CALL allocate_cv(yd_cv, ksmax=yd_cvin%nsmax, kmsmax=yd_cvin%nmsmax, kserr=yd_cvin%nserr, &
               & kparam=yd_cvin%npara, &
               & ldinit=llinit, ldmerr=llmerr, ldtovs=lltovs, ldlam=lllam, &
               & ks3d=yd_cvin%ns3d, ks2d=yd_cvin%ns2d, ks1d=yd_cvin%ns1d)

END SUBROUTINE allocate_cv_cv

! ------------------------------------------------------------------

SUBROUTINE deallocate_ctlvec(yd_cv)
TYPE(control_vector), INTENT(INOUT) :: yd_cv

IF (ASSOCIATED(yd_cv%initcv)) NULLIFY(yd_cv%initcv)
IF (ASSOCIATED(yd_cv%moderr)) NULLIFY(yd_cv%moderr)
IF (ASSOCIATED(yd_cv%tovscv)) NULLIFY(yd_cv%tovscv)
IF (ASSOCIATED(yd_cv%lamcv )) NULLIFY(yd_cv%lamcv )
IF (ASSOCIATED(yd_cv%params)) NULLIFY(yd_cv%params)
IF (ASSOCIATED(yd_cv%data)) DEALLOCATE(yd_cv%data)
IF (ASSOCIATED(yd_cv%ibgn_scale)) DEALLOCATE(yd_cv%ibgn_scale)
yd_cv%lalloc = .false.

END SUBROUTINE deallocate_ctlvec

! ------------------------------------------------------------------

SUBROUTINE assign_cv_cv(yd_cv1,yd_cv2)
TYPE(control_vector), INTENT(INOUT) :: yd_cv1
TYPE(control_vector), INTENT(IN)    :: yd_cv2
TYPE(spectral_field) :: yl_zspec1, yl_zspec2
LOGICAL :: ll_same_init, ll_same_moderr
INTEGER(KIND=JPIM) :: ilen

CALL check(yd_cv1)
CALL check(yd_cv2)
IF (yd_cv1.NEQV.yd_cv2) THEN
  CALL ABOR1('Control vectors are incompatible')
ENDIF

ll_same_init = .true.
IF (yd_cv1%linitcv) THEN
  IF (yd_cv1%lwaveletjb) THEN
    ll_same_init = yd_cv1%nscales==yd_cv2%nscales .AND. (yd_cv1%nserr==yd_cv2%nserr)
  ELSE
    ll_same_init = yd_cv1%nsmax==yd_cv2%nsmax .AND. (yd_cv1%nserr==yd_cv2%nserr)
  ENDIF
ENDIF

IF (LELAM) THEN
  ll_same_init = ll_same_init .AND. (yd_cv1%nmsmax==yd_cv2%nmsmax)
ENDIF

IF (yd_cv1%lmoderr) THEN
  ll_same_moderr = (yd_cv1%nserr==yd_cv2%nserr)
ELSE
  ll_same_moderr = .TRUE.
ENDIF

IF (ll_same_init .AND. ll_same_moderr) THEN
  yd_cv1%data = yd_cv2%data
ELSE
  IF (yd_cv1%linitcv) THEN
    IF (ll_same_init) THEN
      yd_cv1%initcv(:) = yd_cv2%initcv(:)
    ELSEIF (yd_cv1%lwaveletjb) THEN

!--- NB: The finest scale of the lower-resolution control vector (yd_cv2) has a different
!--- grid than the corresponding scale of yd_cv1. Ideally, we should use the spherical
!--- transform to re-grid this scale. But, for now we just zero it in yd_cv1.

      ilen=yd_cv1%ibgn_scale(yd_cv1%nscales-yd_cv2%nscales+2)-1
      yd_cv1%initcv(1:ilen)  = 0.0_JPRB
      IF (SIZE(yd_cv1%initcv(ilen+1:))/=SIZE(yd_cv2%initcv(yd_cv2%ibgn_scale(2):))) THEN
        write (0,*) 'size mismatch: ',SIZE(yd_cv1%initcv(ilen+1:)), &
                   &                  SIZE(yd_cv2%initcv(yd_cv2%ibgn_scale(2):))
        write (0,*) 'yd_cv1%ibgn_scale=',yd_cv1%ibgn_scale
        write (0,*) 'yd_cv2%ibgn_scale=',yd_cv2%ibgn_scale
        CALL ABOR1 ('size mismatch')
      ENDIF
      yd_cv1%initcv(ilen+1:) = yd_cv2%initcv(yd_cv2%ibgn_scale(2):)

    ELSE
      CALL allocate_spec (yl_zspec1,yd_cv1%nsmax,yd_cv1%nmsmax,yd_cv1%ns3d,yd_cv1%ns2d,yd_cv1%ns1d,ngrb)
      CALL allocate_spec (yl_zspec2,yd_cv2%nsmax,yd_cv2%nmsmax,yd_cv2%ns3d,yd_cv2%ns2d,yd_cv2%ns1d,ngrb)
      yl_zspec2=yd_cv2%initcv(:)
      yl_zspec1=yl_zspec2  ! spectral_field type does change of resolution
      yd_cv1%initcv(:)=yl_zspec1
      CALL deallocate_spec(yl_zspec1)
      CALL deallocate_spec(yl_zspec2)
    ENDIF
  ENDIF
  IF (yd_cv1%lmoderr) THEN
    IF (ll_same_moderr) THEN
      yd_cv1%moderr(:) = yd_cv2%moderr(:)
    ELSE
      CALL allocate_spec (yl_zspec1,yd_cv1%nserr,yd_cv1%nserr,yd_cv1%ns3d,yd_cv1%ns2d,yd_cv1%ns1d, &
                        & ngrb)
      CALL allocate_spec (yl_zspec2,yd_cv2%nserr,yd_cv2%nserr,yd_cv2%ns3d,yd_cv2%ns2d,yd_cv2%ns1d, &
                        & ngrb)
      yl_zspec2=yd_cv2%moderr(:)
      yl_zspec1=yl_zspec2  ! spectral_field type does change of resolution
      yd_cv1%moderr(:)=yl_zspec1
      CALL deallocate_spec(yl_zspec1)
      CALL deallocate_spec(yl_zspec2)
    ENDIF
  ENDIF
  IF (yd_cv1%ltovscv) THEN
    yd_cv1%tovscv(:) = yd_cv2%tovscv(:)
  ENDIF
  IF (yd_cv1%lam1d) THEN
    yd_cv1%lamcv(:) = yd_cv2%lamcv(:)
  ENDIF
  IF (yd_cv1%npara>0) THEN
    yd_cv1%params(:) = yd_cv2%params(:)
  ENDIF
ENDIF

END SUBROUTINE assign_cv_cv

! ------------------------------------------------------------------

SUBROUTINE assign_scalar_cv(yd_cv,pscalar)
TYPE(control_vector), INTENT(INOUT) :: yd_cv
REAL(KIND=JPRB), INTENT(IN) :: pscalar

CALL check(yd_cv)
yd_cv%data = pscalar

END SUBROUTINE assign_scalar_cv

! ------------------------------------------------------------------

SUBROUTINE assign_ar_cv(yd_cv,pvec)
TYPE(control_vector), INTENT(INOUT) :: yd_cv
REAL(KIND=JPRB), INTENT(IN) :: pvec(:)
INTEGER(KIND=JPIM) :: ilen, ioff, jf, jj, jl, il, ibgn, iend, igptotg, jscale, ja, jb
TYPE(SPECTRAL_FIELD) :: yl_zfld

CALL check(yd_cv)

IF (SIZE(pvec) == yd_cv%nsizel) then
  yd_cv%data = pvec
ELSEIF (SIZE(pvec) == yd_cv%nsizeg) then
  IF (yd_cv%linitcv) THEN
    IF (.not.yd_cv%lwaveletjb) THEN
      CALL allocate_spec (yl_zfld,yd_cv%nsmax,yd_cv%nmsmax,yd_cv%ns3d,yd_cv%ns2d,yd_cv%ns1d, &
                        & ngrb)
      ilen=yl_zfld%nspsizeg
      yl_zfld=pvec(1:ilen)
      yd_cv%initcv(:)=yl_zfld
      ioff=ilen
      CALL deallocate_spec(yl_zfld)
    ELSE
      igptotg=0
      DO jscale=1,yd_cv%nscales
        DO ja=1,NPRGPNS
          DO jb=1,NPRGPEW
            if (ja==myseta .and. jb==mysetb) then
              ibgn=yd_cv%ibgn_scale(jscale)
              ilen=nproma*(nflevg*yd_cv%ns3d+yd_cv%ns2d)*nblocksl(jscale+NSCALES-yd_cv%nscales)
              iend=ibgn+ilen-1
              yd_cv%initcv(ibgn:iend) = pvec(igptotg+1:igptotg+ilen)
            endif
            igptotg=igptotg+ngptotl(jscale+NSCALES-yd_cv%nscales,ja,jb)
          ENDDO
        ENDDO
      ENDDO
      if (igptotg /= sum(ngptotg(1+NSCALES-yd_cv%nscales:))) &
        & CALL ABOR1 ('assign_ar_cv: indexing problem')
      ioff = igptotg
    ENDIF
  ENDIF
  IF (yd_cv%lmoderr) THEN
    CALL allocate_spec(yl_zfld,yd_cv%nserr,yd_cv%nserr,yd_cv%ns3d,yd_cv%ns2d,yd_cv%ns1d,ngrb)
    ilen=yl_zfld%nspsizeg
    yl_zfld=pvec(ioff+1:ioff+ilen)
    yd_cv%moderr(:)=yl_zfld
    ioff=ioff+ilen
    CALL deallocate_spec(yl_zfld)
  ENDIF
  IF (yd_cv%ltovscv) THEN
    jl=0
    DO jf=1,NVATOVV
      DO jj=1,NOBTOV
        jl=jl+1
        yd_cv%tovscv(jl)=pvec(ioff+mytovoff+jj)
      ENDDO
      ioff=ioff+NOBTOVG
    ENDDO
  ENDIF
  IF (yd_cv%lam1d) THEN
    jj=0
    DO jf=1,yd_cv%ns1d
      DO jl=1,NFLEVL
        il=MYLEVS(jl)
        jj=jj+1
        yd_cv%lamcv(jj)=pvec(ioff+il)
      ENDDO
      ioff=ioff+NFLEVG
    ENDDO
  ENDIF
  IF (yd_cv%npara>0) THEN
    DO jf=1,yd_cv%npara
      yd_cv%params(jf)=pvec(ioff+jf)
    ENDDO
    ioff=ioff+yd_cv%npara
  ENDIF
ELSE 
  CALL ABOR1('Error: cannot assign array to control vector')
ENDIF

END SUBROUTINE assign_ar_cv

! ------------------------------------------------------------------

SUBROUTINE assign_cv_ar(pvec,yd_cv)
REAL(KIND=JPRB), INTENT(OUT) :: pvec(:)
TYPE(control_vector), INTENT(IN) :: yd_cv

CALL check(yd_cv)
IF (SIZE(pvec) == yd_cv%nsizel) then
  pvec = yd_cv%data
ELSEIF (SIZE(pvec) == yd_cv%nsizeg) then
!  This case has to come after the preceeding one so that it isn't
!  executed if running on one processor.
  CALL ABOR1('Assign vector to global array would be very expensive')
ELSE 
  CALL ABOR1('Error: cannot assign control vector to array')
ENDIF

END SUBROUTINE assign_cv_ar

! ------------------------------------------------------------------

SUBROUTINE inverse_cv(yd_cv)
TYPE(control_vector), INTENT(INOUT) :: yd_cv

! divide_cv fails on zero, inverse sets output to zero.
CALL check(yd_cv)
WHERE (yd_cv%data /= 0.0_JPRB)
 yd_cv%data = 1.0_JPRB/yd_cv%data
ENDWHERE

END SUBROUTINE inverse_cv

! ------------------------------------------------------------------

SUBROUTINE inverse_cv_cv(yd_cvin,yd_cvout)
TYPE(control_vector), INTENT(IN)    :: yd_cvin
TYPE(control_vector), INTENT(INOUT) :: yd_cvout

! divide_cv fails on zero, inverse sets output to zero.
CALL check(yd_cvin,yd_cvout)

WHERE (yd_cvin%data == 0.0_JPRB)
 yd_cvout%data = 0.0_JPRB
ELSEWHERE
 yd_cvout%data = 1.0_JPRB/yd_cvin%data
ENDWHERE

END SUBROUTINE inverse_cv_cv

! ------------------------------------------------------------------

SUBROUTINE check_cv(yd_cv)
TYPE(control_vector), INTENT(IN) :: yd_cv

IF (.NOT.(yd_cv%lalloc)) THEN
  CALL ABOR1('Control vector not allocated')
ENDIF
!yt the following is temporary, lalloc can be removed after
!yt this has been tested.
IF (ASSOCIATED(yd_cv%data) .AND. .NOT.(yd_cv%lalloc)) THEN
  CALL ABOR1('Control vector associated but not lalloc')
ENDIF
IF (.NOT.ASSOCIATED(yd_cv%data) .AND. yd_cv%lalloc) THEN
  CALL ABOR1('Control vector lalloc but not associated')
ENDIF

END SUBROUTINE check_cv

! ------------------------------------------------------------------

SUBROUTINE check_cv_cv(yd_cv1,yd_cv2)
TYPE(control_vector), INTENT(IN) :: yd_cv1, yd_cv2

IF (.NOT.(yd_cv1%lalloc .AND. yd_cv2%lalloc)) THEN
  CALL ABOR1('Control vectors not allocated')
ENDIF
IF (yd_cv1.NEQV.yd_cv2) THEN
  CALL ABOR1('Control vectors are incompatible')
ENDIF

IF ((yd_cv1%nsmax/=yd_cv2%nsmax) .OR. (yd_cv1%nserr/=yd_cv2%nserr) .OR. &
 &((yd_cv1%nmsmax/=yd_cv2%nmsmax).AND.(LELAM))) THEN
  CALL ABOR1('Control vectors have different resolution')
ENDIF

END SUBROUTINE check_cv_cv

! ------------------------------------------------------------------

LOGICAL FUNCTION equiv_cv(yd_cv1,yd_cv2)
TYPE(control_vector), INTENT(IN) :: yd_cv1
TYPE(control_vector), INTENT(IN) :: yd_cv2
LOGICAL :: ll

ll = .true.
ll = ll .AND. (yd_cv1%linitcv.EQV.yd_cv2%linitcv)
ll = ll .AND. (yd_cv1%lmoderr.EQV.yd_cv2%lmoderr)
ll = ll .AND. (yd_cv1%ltovscv.EQV.yd_cv2%ltovscv)
ll = ll .AND. (yd_cv1%lam1d  .EQV.yd_cv2%lam1d  )

ll = ll .AND. (yd_cv1%ns2d == yd_cv2%ns2d)
ll = ll .AND. (yd_cv1%ns3d == yd_cv2%ns3d)
ll = ll .AND. (yd_cv1%ns1d == yd_cv2%ns1d)
ll = ll .AND. (yd_cv1%npara== yd_cv2%npara)

equiv_cv=ll
END FUNCTION equiv_cv
  
! ------------------------------------------------------------------
  
LOGICAL FUNCTION nequiv_cv(yd_cv1,yd_cv2)
TYPE(control_vector), INTENT(IN) :: yd_cv1
TYPE(control_vector), INTENT(IN) :: yd_cv2
  
nequiv_cv = .NOT. equiv_cv(yd_cv1,yd_cv2)
  
END FUNCTION nequiv_cv
    
! ------------------------------------------------------------------

REAL(KIND=JPRB) FUNCTION dot_product_cv_cv(yd_cv1,yd_cv2)
TYPE (control_vector), INTENT(IN) :: yd_cv1,yd_cv2
REAL(KIND=JPRB) :: zz1,zz2,zz3,zz4,zz5

CALL check(yd_cv1,yd_cv2)
CALL dot_product_ctlvec(yd_cv1,yd_cv2,zz1,zz2,zz3,zz4,zz5,.false.)

dot_product_cv_cv=zz1+zz2+zz3+zz4+zz5

END FUNCTION dot_product_cv_cv

! ------------------------------------------------------------------

REAL(KIND=JPRB) FUNCTION dot_product_weight(yd_cv1,yd_cv2,yd_cvw)
TYPE (control_vector), INTENT(IN) :: yd_cv1,yd_cv2,yd_cvw
REAL(KIND=JPRB) :: zz1,zz2,zz3,zz4,zz5
TYPE (control_vector) :: yl_zcv

CALL GSTATS(1933,0)
CALL check(yd_cv1,yd_cvw)
CALL check(yd_cv2,yd_cvw)

CALL ALLOCATE_CTLVEC(yl_zcv,yd_cvw)

yl_zcv%data = yd_cv1%data * yd_cv2%data
CALL GSTATS(1933,1)
CALL dot_product_ctlvec(yl_zcv,yd_cvw,zz1,zz2,zz3,zz4,zz5,.false.)

CALL DEALLOCATE_CTLVEC(yl_zcv)

dot_product_weight=zz1+zz2+zz3+zz4+zz5

END FUNCTION dot_product_weight

! ------------------------------------------------------------------

SUBROUTINE cv_sqnorm1(yd_cv,pnorm)
TYPE (control_vector), INTENT(IN) :: yd_cv
REAL(KIND=JPRB), INTENT(OUT) :: pnorm
REAL(KIND=JPRB) :: zz1,zz2,zz3,zz4,zz5

CALL check(yd_cv)
CALL dot_product_ctlvec(yd_cv,yd_cv,zz1,zz2,zz3,zz4,zz5,.true.)

pnorm=zz1+zz2+zz3+zz4+zz5

END SUBROUTINE cv_sqnorm1

! ------------------------------------------------------------------

SUBROUTINE cv_sqnorm4(yd_cv,pnorm1,pnorm2,pnorm3,pnorm4,pnorm5)
TYPE (control_vector), INTENT(IN) :: yd_cv
REAL(KIND=JPRB), INTENT(OUT) :: pnorm1,pnorm2,pnorm3,pnorm4,pnorm5

CALL check(yd_cv)
CALL dot_product_ctlvec(yd_cv,yd_cv,pnorm1,pnorm2,pnorm3,pnorm4,pnorm5,.true.)

END SUBROUTINE cv_sqnorm4

! ------------------------------------------------------------------

REAL(KIND=JPRB) FUNCTION ctlvec_norm(yd_cv)
TYPE (control_vector), INTENT(IN) :: yd_cv
REAL(KIND=JPRB) :: zz

CALL cv_sqnorm1(yd_cv,zz)
ctlvec_norm=SQRT(zz)

END FUNCTION ctlvec_norm

! ------------------------------------------------------------------

REAL(KIND=JPRB) FUNCTION maxval_cv(yd_cv)
TYPE (control_vector), INTENT(IN) :: yd_cv
REAL(KIND=JPRB) :: zz

CALL GSTATS(256,0)
zz = MAXVAL(yd_cv%data)
CALL MPL_ALLREDUCE(zz,'MAX',CDSTRING='CONTROL_VECTORS:MAXVAL_CV')
maxval_cv=zz
CALL GSTATS(256,1)

END FUNCTION maxval_cv

! ------------------------------------------------------------------

REAL(KIND=JPRB) FUNCTION minval_cv(yd_cv)
TYPE (control_vector), INTENT(IN) :: yd_cv
REAL(KIND=JPRB) :: zz

CALL GSTATS(256,0)
zz = MINVAL(yd_cv%data)
CALL MPL_ALLREDUCE(zz,'MIN',CDSTRING='CONTROL_VECTORS:MINVAL_CV')
minval_cv=zz
CALL GSTATS(256,1)

END FUNCTION minval_cv

! ------------------------------------------------------------------

REAL(KIND=JPRB) FUNCTION sum_cv(yd_cv)
TYPE (control_vector), INTENT(IN) :: yd_cv
REAL(KIND=JPRB) :: zz

! At the moment, this function is only used to print diagnostics in
! m1qn3 if impres>=5 which is not the case in the IFS by default.
! It is NOT reproducible.

CALL GSTATS(256,0)
zz=SUM(yd_cv%data(:))
CALL MPL_ALLREDUCE(zz,'SUM',LDREPROD=.false.,CDSTRING='CONTROL_VECTORS:SUM_CV')
sum_cv = zz
CALL GSTATS(256,1)

END FUNCTION sum_cv

! ------------------------------------------------------------------

END MODULE CONTROL_VECTORS
